"""
Defines all types in the variable-free language for the WikiTablesQuestions dataset defined in the
following paper by Liang et al. (2018)
Memory Augmented Policy Optimization for Program Synthesis with Generalization.
"""


from allennlp.semparse.type_declarations.type_declaration import (NamedBasicType, ComplexType,
                                                                  MultiMatchNamedBasicType,
                                                                  NameMapper)


# Basic types
# Note that there aren't cell types and part types in this gramamr. They are all of type string. The
# constants will still come from the `TableQuestionKnowledgeGraph`, and are identified as parts and
# cells, and the executor can process them accordingly, but the grammar does not differentiate
# between them. Also note that columns are basic types in this grammar.
ROW_TYPE = NamedBasicType("ROW")
# The following three type signatures are assigned such that they're easy to understand from the
# first letters, while being different from the cell types.
DATE_COLUMN_TYPE = NamedBasicType("MCOLUMN")  # M for month
NUMBER_COLUMN_TYPE = NamedBasicType("FCOLUMN")  # F for float
STRING_COLUMN_TYPE = NamedBasicType("TCOLUMN")  # T for token
GENERIC_COLUMN_TYPE = MultiMatchNamedBasicType("GCOLUMN", [STRING_COLUMN_TYPE, DATE_COLUMN_TYPE,
                                                           NUMBER_COLUMN_TYPE])
COMPARABLE_COLUMN_TYPE = MultiMatchNamedBasicType("CCOLUMN", [NUMBER_COLUMN_TYPE, DATE_COLUMN_TYPE])

NUMBER_TYPE = NamedBasicType("NUMBER")
DATE_TYPE = NamedBasicType("DATE")
STRING_TYPE = NamedBasicType("STRING")

# We start with just the generic column type, and add the specific column types to the set only if
# we see the corresponding types in the table.
BASIC_TYPES = {ROW_TYPE, GENERIC_COLUMN_TYPE, NUMBER_TYPE, DATE_TYPE, STRING_TYPE}
STARTING_TYPES = {NUMBER_TYPE, DATE_TYPE, STRING_TYPE}

# Complex types
# Type for selecting the value in a column in a set of rows. "select" and "mode" functions.
SELECT_TYPE = ComplexType(ROW_TYPE, ComplexType(GENERIC_COLUMN_TYPE, STRING_TYPE))

# Type for filtering rows given a column. "argmax", "argmin" and "same_as" (select all rows with the
# same value under the given column as the given row does under the given column). While "same_as"
# takes any column, "argmax" and "argmin" take only comparable columns (i.e. dates or numbers).
# Note that the values used for comparison in "argmax" and "argmin" can only come from column
# lookups in this language. In LambdaDCS, there's a lambda function that is applied to the rows to
# get the values, but here, we simply have a column name.
ROW_FILTER_WITH_GENERIC_COLUMN = ComplexType(ROW_TYPE, ComplexType(GENERIC_COLUMN_TYPE, ROW_TYPE))
ROW_FILTER_WITH_COMPARABLE_COLUMN = ComplexType(ROW_TYPE, ComplexType(COMPARABLE_COLUMN_TYPE, ROW_TYPE))

# "filter_number_greater", "filter_number_equals" etc.
ROW_FILTER_WITH_COLUMN_AND_NUMBER = ComplexType(ROW_TYPE,
                                                ComplexType(NUMBER_COLUMN_TYPE,
                                                            ComplexType(NUMBER_TYPE, ROW_TYPE)))

# "filter_date_greater", "filter_date_equals" etc.
ROW_FILTER_WITH_COLUMN_AND_DATE = ComplexType(ROW_TYPE,
                                              ComplexType(DATE_COLUMN_TYPE,
                                                          ComplexType(DATE_TYPE, ROW_TYPE)))

# "filter_in" and "filter_not_in"
ROW_FILTER_WITH_COLUMN_AND_STRING = ComplexType(ROW_TYPE,
                                                ComplexType(STRING_COLUMN_TYPE,
                                                            ComplexType(STRING_TYPE, ROW_TYPE)))

ROW_FILTER = ComplexType(ROW_TYPE, ROW_TYPE)  # first, last, previous, next etc.

# This language lets you count only rows!
COUNT_TYPE = ComplexType(ROW_TYPE, NUMBER_TYPE)

# Numerical operations on numbers in the given column. "max", "min", "sum", "average" etc.
ROW_NUM_OP = ComplexType(ROW_TYPE, ComplexType(NUMBER_COLUMN_TYPE, NUMBER_TYPE))

# Numerical difference within the given column.
NUM_DIFF_WITH_COLUMN = ComplexType(ROW_TYPE, ComplexType(ROW_TYPE, ComplexType(NUMBER_COLUMN_TYPE,
                                                                               NUMBER_TYPE)))

# Date function takes three numbers and makes a date
DATE_FUNCTION_TYPE = ComplexType(NUMBER_TYPE, ComplexType(NUMBER_TYPE, ComplexType(NUMBER_TYPE,
                                                                                   DATE_TYPE)))


# We have one name mapper for all generic functions, and different ones for each of the column types
# because we will add those to the final name mapping only if needed, based on the table content.
generic_name_mapper = NameMapper()  # pylint: disable=invalid-name
string_column_name_mapper = NameMapper(alias_prefix="S")  # pylint: disable=invalid-name
number_column_name_mapper = NameMapper(alias_prefix="N")  # pylint: disable=invalid-name
date_column_name_mapper = NameMapper(alias_prefix="D")  # pylint: disable=invalid-name
comparable_column_name_mapper = NameMapper(alias_prefix="O")  # pylint: disable=invalid-name

generic_name_mapper.map_name_with_signature("all_rows", ROW_TYPE)

# <r,<g,s>>
generic_name_mapper.map_name_with_signature("select", SELECT_TYPE)
generic_name_mapper.map_name_with_signature("mode", SELECT_TYPE)

# <r,<c,r>>
comparable_column_name_mapper.map_name_with_signature("argmax", ROW_FILTER_WITH_COMPARABLE_COLUMN)
comparable_column_name_mapper.map_name_with_signature("argmin", ROW_FILTER_WITH_COMPARABLE_COLUMN)

# <r,<g,r>>
generic_name_mapper.map_name_with_signature("same_as", ROW_FILTER_WITH_GENERIC_COLUMN)

# <r,<f,<n,r>>>
number_column_name_mapper.map_name_with_signature("filter_number_greater", ROW_FILTER_WITH_COLUMN_AND_NUMBER)
number_column_name_mapper.map_name_with_signature("filter_number_greater_equals",
                                                  ROW_FILTER_WITH_COLUMN_AND_NUMBER)
number_column_name_mapper.map_name_with_signature("filter_number_lesser", ROW_FILTER_WITH_COLUMN_AND_NUMBER)
number_column_name_mapper.map_name_with_signature("filter_number_lesser_equals", ROW_FILTER_WITH_COLUMN_AND_NUMBER)
number_column_name_mapper.map_name_with_signature("filter_number_equals", ROW_FILTER_WITH_COLUMN_AND_NUMBER)
number_column_name_mapper.map_name_with_signature("filter_number_not_equals", ROW_FILTER_WITH_COLUMN_AND_NUMBER)

# <r,<y,<d,r>>>
date_column_name_mapper.map_name_with_signature("filter_date_greater", ROW_FILTER_WITH_COLUMN_AND_DATE)
date_column_name_mapper.map_name_with_signature("filter_date_greater_equals", ROW_FILTER_WITH_COLUMN_AND_DATE)
date_column_name_mapper.map_name_with_signature("filter_date_lesser", ROW_FILTER_WITH_COLUMN_AND_DATE)
date_column_name_mapper.map_name_with_signature("filter_date_lesser_equals", ROW_FILTER_WITH_COLUMN_AND_DATE)
date_column_name_mapper.map_name_with_signature("filter_date_equals", ROW_FILTER_WITH_COLUMN_AND_DATE)
date_column_name_mapper.map_name_with_signature("filter_date_not_equals", ROW_FILTER_WITH_COLUMN_AND_DATE)

# <r,<t,<s,r>>>
string_column_name_mapper.map_name_with_signature("filter_in", ROW_FILTER_WITH_COLUMN_AND_STRING)
string_column_name_mapper.map_name_with_signature("filter_not_in", ROW_FILTER_WITH_COLUMN_AND_STRING)

# <r,r>
generic_name_mapper.map_name_with_signature("first", ROW_FILTER)
generic_name_mapper.map_name_with_signature("last", ROW_FILTER)
generic_name_mapper.map_name_with_signature("previous", ROW_FILTER)
generic_name_mapper.map_name_with_signature("next", ROW_FILTER)

# <r,n>
generic_name_mapper.map_name_with_signature("count", COUNT_TYPE)

# <r,<f,n>>
number_column_name_mapper.map_name_with_signature("max", ROW_NUM_OP)
number_column_name_mapper.map_name_with_signature("min", ROW_NUM_OP)
number_column_name_mapper.map_name_with_signature("average", ROW_NUM_OP)
number_column_name_mapper.map_name_with_signature("sum", ROW_NUM_OP)

# <r,<r,<f,n>>>
number_column_name_mapper.map_name_with_signature("diff", NUM_DIFF_WITH_COLUMN)

# <n,<n,<n,d>>>
generic_name_mapper.map_name_with_signature("date", DATE_FUNCTION_TYPE)

COMMON_NAME_MAPPING = generic_name_mapper.common_name_mapping
COMMON_TYPE_SIGNATURE = generic_name_mapper.common_type_signature
STRING_COLUMN_NAME_MAPPING = string_column_name_mapper.common_name_mapping
STRING_COLUMN_TYPE_SIGNATURE = string_column_name_mapper.common_type_signature
NUMBER_COLUMN_NAME_MAPPING = number_column_name_mapper.common_name_mapping
NUMBER_COLUMN_TYPE_SIGNATURE = number_column_name_mapper.common_type_signature
DATE_COLUMN_NAME_MAPPING = date_column_name_mapper.common_name_mapping
DATE_COLUMN_TYPE_SIGNATURE = date_column_name_mapper.common_type_signature
COMPARABLE_COLUMN_NAME_MAPPING = comparable_column_name_mapper.common_name_mapping
COMPARABLE_COLUMN_TYPE_SIGNATURE = comparable_column_name_mapper.common_type_signature



# pylint: disable=invalid-name,no-self-use
import argparse
import json

from flaky import flaky

from allennlp.commands.evaluate import evaluate_from_args, Evaluate
from allennlp.common.testing import AllenNlpTestCase


class TestEvaluate(AllenNlpTestCase):
    def setUp(self):
        super().setUp()

        self.parser = argparse.ArgumentParser(description="Testing")
        subparsers = self.parser.add_subparsers(title='Commands', metavar='')
        Evaluate().add_subparser('evaluate', subparsers)

    @flaky
    def test_evaluate_from_args(self):
        kebab_args = ["evaluate", str(self.FIXTURES_ROOT / "bidaf" / "serialization" / "model.tar.gz"),
                      str(self.FIXTURES_ROOT / "data" / "squad.json"),
                      "--cuda-device", "-1"]

        args = self.parser.parse_args(kebab_args)
        metrics = evaluate_from_args(args)
        assert metrics.keys() == {'span_acc', 'end_acc', 'start_acc', 'em', 'f1', 'loss'}

    def test_output_file_evaluate_from_args(self):
        output_file = str(self.TEST_DIR / "metrics.json")
        kebab_args = ["evaluate", str(self.FIXTURES_ROOT / "bidaf" / "serialization" / "model.tar.gz"),
                      str(self.FIXTURES_ROOT / "data" / "squad.json"),
                      "--cuda-device", "-1",
                      "--output-file", output_file]
        args = self.parser.parse_args(kebab_args)
        computed_metrics = evaluate_from_args(args)
        with open(output_file, 'r') as file:
            saved_metrics = json.load(file)
        assert computed_metrics == saved_metrics



from allennlp.semparse.worlds.nlvr_world import NlvrWorld
from allennlp.semparse.worlds.wikitables_world import WikiTablesWorld
from allennlp.semparse.worlds.wikitables_variable_free_world import WikiTablesVariableFreeWorld
from allennlp.semparse.worlds.atis_world import AtisWorld



# pylint: disable=no-self-use,invalid-name
from collections import defaultdict

from allennlp.common.testing import AllenNlpTestCase
from allennlp.data import Token, Vocabulary
from allennlp.data.token_indexers import TokenCharactersIndexer
from allennlp.data.tokenizers.character_tokenizer import CharacterTokenizer


class CharacterTokenIndexerTest(AllenNlpTestCase):
    def test_count_vocab_items_respects_casing(self):
        indexer = TokenCharactersIndexer("characters")
        counter = defaultdict(lambda: defaultdict(int))
        indexer.count_vocab_items(Token("Hello"), counter)
        indexer.count_vocab_items(Token("hello"), counter)
        assert counter["characters"] == {"h": 1, "H": 1, "e": 2, "l": 4, "o": 2}

        indexer = TokenCharactersIndexer("characters", CharacterTokenizer(lowercase_characters=True))
        counter = defaultdict(lambda: defaultdict(int))
        indexer.count_vocab_items(Token("Hello"), counter)
        indexer.count_vocab_items(Token("hello"), counter)
        assert counter["characters"] == {"h": 2, "e": 2, "l": 4, "o": 2}

    def test_as_array_produces_token_sequence(self):
        indexer = TokenCharactersIndexer("characters")
        padded_tokens = indexer.pad_token_sequence({'k': [[1, 2, 3, 4, 5], [1, 2, 3], [1]]},
                                                   desired_num_tokens={'k': 4},
                                                   padding_lengths={"num_token_characters": 10})
        assert padded_tokens == {'k': [[1, 2, 3, 4, 5, 0, 0, 0, 0, 0],
                                       [1, 2, 3, 0, 0, 0, 0, 0, 0, 0],
                                       [1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                       [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]}

    def test_tokens_to_indices_produces_correct_characters(self):
        vocab = Vocabulary()
        vocab.add_token_to_namespace("A", namespace='characters')
        vocab.add_token_to_namespace("s", namespace='characters')
        vocab.add_token_to_namespace("e", namespace='characters')
        vocab.add_token_to_namespace("n", namespace='characters')
        vocab.add_token_to_namespace("t", namespace='characters')
        vocab.add_token_to_namespace("c", namespace='characters')

        indexer = TokenCharactersIndexer("characters")
        indices = indexer.tokens_to_indices([Token("sentential")], vocab, "char")
        assert indices == {"char": [[3, 4, 5, 6, 4, 5, 6, 1, 1, 1]]}

    def test_start_and_end_tokens(self):
        vocab = Vocabulary()
        vocab.add_token_to_namespace("A", namespace='characters')  # 2
        vocab.add_token_to_namespace("s", namespace='characters')  # 3
        vocab.add_token_to_namespace("e", namespace='characters')  # 4
        vocab.add_token_to_namespace("n", namespace='characters')  # 5
        vocab.add_token_to_namespace("t", namespace='characters')  # 6
        vocab.add_token_to_namespace("c", namespace='characters')  # 7
        vocab.add_token_to_namespace("<", namespace='characters')  # 8
        vocab.add_token_to_namespace(">", namespace='characters')  # 9
        vocab.add_token_to_namespace("/", namespace='characters')  # 10

        indexer = TokenCharactersIndexer("characters", start_tokens=["<s>"], end_tokens=["</s>"])
        indices = indexer.tokens_to_indices([Token("sentential")], vocab, "char")
        assert indices == {"char": [[8, 3, 9],
                                    [3, 4, 5, 6, 4, 5, 6, 1, 1, 1],
                                    [8, 10, 3, 9]]}

    def test_min_padding_length(self):
        sentence = "AllenNLP is awesome ."
        tokens = [Token(token) for token in sentence.split(" ")]
        vocab = Vocabulary()
        vocab.add_token_to_namespace("A", namespace="characters")  # 2
        vocab.add_token_to_namespace("l", namespace="characters")  # 3
        vocab.add_token_to_namespace("e", namespace="characters")  # 4
        vocab.add_token_to_namespace("n", namespace="characters")  # 5
        vocab.add_token_to_namespace("N", namespace="characters")  # 6
        vocab.add_token_to_namespace("L", namespace="characters")  # 7
        vocab.add_token_to_namespace("P", namespace="characters")  # 8
        vocab.add_token_to_namespace("i", namespace="characters")  # 9
        vocab.add_token_to_namespace("s", namespace="characters")  # 10
        vocab.add_token_to_namespace("a", namespace="characters")  # 11
        vocab.add_token_to_namespace("w", namespace="characters")  # 12
        vocab.add_token_to_namespace("o", namespace="characters")  # 13
        vocab.add_token_to_namespace("m", namespace="characters")  # 14
        vocab.add_token_to_namespace(".", namespace="characters")  # 15

        indexer = TokenCharactersIndexer("characters", min_padding_length=10)
        indices = indexer.tokens_to_indices(tokens, vocab, "char")
        key_padding_lengths = "num_token_characters"
        value_padding_lengths = 0
        for token in indices["char"]:
            item = indexer.get_padding_lengths(token)
            value = item.values()
            value_padding_lengths = max(value_padding_lengths, max(value))
        padded = indexer.pad_token_sequence(indices,
                                            {"char": len(indices["char"])},
                                            {key_padding_lengths: value_padding_lengths})
        assert padded == {"char": [[2, 3, 3, 4, 5, 6, 7, 8, 0, 0],
                                   [9, 10, 0, 0, 0, 0, 0, 0, 0, 0],
                                   [11, 12, 4, 10, 13, 14, 4, 0, 0, 0],
                                   [15, 0, 0, 0, 0, 0, 0, 0, 0, 0]]}



from typing import List, Dict, Tuple, Union, Any
from collections import defaultdict
import logging

from allennlp.semparse import util as semparse_util
from allennlp.semparse.worlds.world import ExecutionError
from allennlp.semparse.contexts.table_question_knowledge_graph import MONTH_NUMBERS
from allennlp.semparse.contexts import TableQuestionContext
from allennlp.tools import wikitables_evaluator as evaluator

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

NestedList = List[Union[str, List]]  # pylint: disable=invalid-name

class Date:
    def __init__(self, year: int, month: int, day: int) -> None:
        self.year = year
        self.month = month
        self.day = day

    def __eq__(self, other) -> bool:
        # Note that the logic below renders equality to be non-transitive. That is,
        # Date(2018, -1, -1) == Date(2018, 2, 3) and Date(2018, -1, -1) == Date(2018, 4, 5)
        # but Date(2018, 2, 3) != Date(2018, 4, 5).
        if not isinstance(other, Date):
            return False
        year_is_same = self.year == -1 or other.year == -1 or self.year == other.year
        month_is_same = self.month == -1 or other.month == -1 or self.month == other.month
        day_is_same = self.day == -1 or other.day == -1 or self.day == other.day
        return year_is_same and month_is_same and day_is_same

    def __gt__(self, other) -> bool:
        # pylint: disable=too-many-return-statements
        # The logic below is tricky, and is based on some assumptions we make about date comparison.
        # Year, month or day being -1 means that we do not know its value. In those cases, the
        # we consider the comparison to be undefined, and return False if all the fields that are
        # more significant than the field being compared are equal. However, when year is -1 for both
        # dates being compared, it is safe to assume that the year is not specified because it is
        # the same. So we make an exception just in that case. That is, we deem the comparison
        # undefined only when one of the year values is -1, but not both.
        if not isinstance(other, Date):
            return False  # comparison undefined
        # We're doing an exclusive or below.
        if (self.year == -1) != (other.year == -1):
            return False  # comparison undefined
        # If both years are -1, we proceed.
        if self.year != other.year:
            return self.year > other.year
        # The years are equal and not -1, or both are -1.
        if self.month == -1 or other.month == -1:
            return False
        if self.month != other.month:
            return self.month > other.month
        # The months and years are equal and not -1
        if self.day == -1 or other.day == -1:
            return False
        return self.day > other.day

    def __ge__(self, other) -> bool:
        if not isinstance(other, Date):
            return False
        return self > other or self == other

    def __str__(self):
        return f"{self.year}-{self.month}-{self.day}"


class WikiTablesVariableFreeExecutor:
    # pylint: disable=too-many-public-methods
    """
    Implements the functions in the variable free language we use, that's inspired by the one in
    "Memory Augmented Policy Optimization for Program Synthesis with Generalization" by Liang et al.

    Parameters
    ----------
    table_data : ``List[Dict[str, str]]``
        All the rows in the table on which the executor will be used. The class expects each row to
        be represented as a dict from column names to corresponding cell values.
    """
    def __init__(self, table_data: List[Dict[str, str]]) -> None:
        self.table_data = table_data

    def __eq__(self, other):
        if not isinstance(other, WikiTablesVariableFreeExecutor):
            return False
        return self.table_data == other.table_data

    def execute(self, logical_form: str) -> Any:
        if not logical_form.startswith("("):
            logical_form = f"({logical_form})"
        logical_form = logical_form.replace(",", " ")
        expression_as_list = semparse_util.lisp_to_nested_expression(logical_form)
        # Expression list has an additional level of
        # nesting at the top. For example, if the
        # logical form is
        # "(select all_rows fb:row.row.league)",
        # the expression list will be
        # [['select', 'all_rows', 'fb:row.row.league']].
        # Removing the top most level of nesting.
        result = self._handle_expression(expression_as_list[0])
        return result

    def evaluate_logical_form(self, logical_form: str, target_list: List[str]) -> bool:
        """
        Takes a logical form, and the list of target values as strings from the original lisp
        string, and returns True iff the logical form executes to the target list.
        """
        normalized_target_list = [TableQuestionContext.normalize_string(value) for value in
                                  target_list]
        target_value_list = evaluator.to_value_list(normalized_target_list)
        try:
            denotation = self.execute(logical_form)
        except ExecutionError:
            logger.warning(f'Failed to execute: {logical_form}')
            return False
        if isinstance(denotation, list):
            denotation_list = [str(denotation_item) for denotation_item in denotation]
        else:
            denotation_list = [str(denotation)]
        denotation_value_list = evaluator.to_value_list(denotation_list)
        return evaluator.check_denotation(target_value_list, denotation_value_list)

    ## Helper functions
    def _handle_expression(self, expression_list):
        if isinstance(expression_list, list) and len(expression_list) == 1:
            expression = expression_list[0]
        else:
            expression = expression_list
        if isinstance(expression, list):
            # This is a function application.
            function_name = expression[0]
        else:
            # This is a constant (like "all_rows" or "2005")
            return self._handle_constant(expression)
        try:
            function = getattr(self, function_name)
            return function(*expression[1:])
        except AttributeError:
            raise ExecutionError(f"Function not found: {function_name}")

    def _handle_constant(self, constant: str) -> Union[List[Dict[str, str]], str, float]:
        if constant == "all_rows":
            return self.table_data
        try:
            return float(constant)
        except ValueError:
            # The constant is not a number. Returning as-is if it is a string.
            if constant.startswith("string:"):
                return constant.replace("string:", "")
            raise ExecutionError(f"Cannot handle constant: {constant}")

    @staticmethod
    def _get_number_row_pairs_to_filter(row_list: List[Dict[str, str]],
                                        column_name: str) -> List[Tuple[float, Dict[str, str]]]:
        """
        Helper method that takes a row list and a column name, and returns a list of tuples, each
        containing as the first element a number taken from that column, and the corresponding row
        as the second element. The output can be used to compare rows based on the numbers.
        """
        if not row_list:
            return []
        try:
            # Various symbols like commas, dollar signs would have been converted to _. Removing
            # them for float conversion.
            cell_row_pairs = [(float(row[column_name].replace('_', '')), row) for row in row_list]
        except ValueError:
            # This means that at least one of the cells is not numerical.
            return []
        return cell_row_pairs

    def _get_date_row_pairs_to_filter(self,
                                      row_list: List[Dict[str, str]],
                                      column_name: str) -> List[Tuple[Date, Dict[str, str]]]:
        """
        Helper method that takes a row list and a column name, and returns a list of tuples, each
        containing as the first element a date taken from that column, and the corresponding row as
        the second element. The output can be used to compare rows based on the dates.
        """
        if not row_list:
            return []
        cell_row_pairs = [(self._make_date(row[column_name]), row) for row in row_list]
        return cell_row_pairs

    @staticmethod
    def _make_date(cell_string: str) -> Date:
        string_parts = cell_string.split("_")
        year = -1
        month = -1
        day = -1
        for part in string_parts:
            if part.isdigit():
                if len(part) == 4:
                    year = int(part)
                else:
                    day = int(part)
            elif part in MONTH_NUMBERS:
                month = MONTH_NUMBERS[part]
        return Date(year, month, day)

    @staticmethod
    def _value_looks_like_date(cell_value: str) -> bool:
        # TODO (pradeep): This will be unnecessary when we have column types identified.
        # We try to figure out if the values being compared are simple numbers or dates. We use
        # simple rules here: that the string contains less than 4 parts, and one of the parts is a
        # month name. Note that this will not consider strings with just years as dates. That's fine
        # because we can compare them as numbers.
        values_are_dates = False
        cell_value_parts = cell_value.split('_')
        # Check if the number of parts in the string are 3 or fewer. If not, it's probably neither a
        # date nor a number.
        if len(cell_value_parts) <= 3:
            for part in cell_value_parts:
                if part in MONTH_NUMBERS:
                    values_are_dates = True
        return values_are_dates

    def _get_row_index(self, row: Dict[str, str]) -> int:
        """
        Takes a row and returns its index in the full list of rows. If the row does not occur in the
        table (which should never happen because this function will only be called with a row that
        is the result of applying one or more functions on the table rows), the method returns -1.
        """
        row_index = -1
        for index, table_row in enumerate(self.table_data):
            if table_row == row:
                row_index = index
                break
        return row_index

    ## Functions in the language
    def select(self, row_expression_list: NestedList, column_name: str) -> List[str]:
        """
        Select function takes a list of rows and a column name and returns a list of cell values as
        strings.
        """
        row_list = self._handle_expression(row_expression_list)
        return [row[column_name] for row in row_list]

    def argmax(self, row_expression_list: NestedList, column_name: str) -> List[Dict[str, str]]:
        """
        Takes a list of rows and a column name and returns a list containing a single row (dict from
        columns to cells) that has the maximum numerical value in the given column. We return a list
        instead of a single dict to be consistent with the return type of `_select` and `_all_rows`.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        # We just check whether the first cell value is a date or number and assume that the rest
        # are the same kind of values.
        first_cell_value = row_list[0][column_name]
        if self._value_looks_like_date(first_cell_value):
            value_row_pairs = self._get_date_row_pairs_to_filter(row_list, column_name)
        else:
            value_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)  # type: ignore
        if not value_row_pairs:
            return []
        # Returns a list containing the row with the max cell value.
        return [sorted(value_row_pairs, key=lambda x: x[0], reverse=True)[0][1]]

    def argmin(self, row_expression_list: NestedList, column_name: str) -> List[Dict[str, str]]:
        """
        Takes a list of rows and a column and returns a list containing a single row (dict from
        columns to cells) that has the minimum numerical value in the given column. We return a list
        instead of a single dict to be consistent with the return type of `_select` and `_all_rows`.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        # We just check whether the first cell value is a date or number and assume that the rest
        # are the same kind of values.
        first_cell_value = row_list[0][column_name]
        if self._value_looks_like_date(first_cell_value):
            value_row_pairs = self._get_date_row_pairs_to_filter(row_list, column_name)
        else:
            value_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)  # type: ignore
        if not value_row_pairs:
            return []
        # Returns a list containing the row with the max cell value.
        return [sorted(value_row_pairs, key=lambda x: x[0])[0][1]]

    def filter_number_greater(self,
                              row_expression_list: NestedList,
                              column_name: str,
                              value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows as an expression, a column, and a numerical value expression and
        returns all the rows where the value in that column is greater than the given value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, float):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value > filter_value:
                return_list.append(row)
        return return_list

    def filter_number_greater_equals(self,
                                     row_expression_list: NestedList,
                                     column_name: str,
                                     value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows as an expression, a column, and a numerical value expression and
        returns all the rows where the value in that column is greater than or equal to the given
        value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, float):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value >= filter_value:
                return_list.append(row)
        return return_list

    def filter_number_lesser(self,
                             row_expression_list: NestedList,
                             column_name: str,
                             value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows as an expression, a column, and a numerical value expression and
        returns all the rows where the value in that column is less than the given value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, float):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value < filter_value:
                return_list.append(row)
        return return_list

    def filter_number_lesser_equals(self,
                                    row_expression_list: NestedList,
                                    column_name: str,
                                    value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows, a column, and a numerical value and returns all the rows where the
        value in that column is lesser than or equal to the given value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, float):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value <= filter_value:
                return_list.append(row)
        return return_list

    def filter_number_equals(self,
                             row_expression_list: NestedList,
                             column_name: str,
                             value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows, a column, and a numerical value and returns all the rows where the
        value in that column equals the given value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, float):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value == filter_value:
                return_list.append(row)
        return return_list

    def filter_number_not_equals(self,
                                 row_expression_list: NestedList,
                                 column_name: str,
                                 value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows, a column, and a numerical value and returns all the rows where the
        value in that column is not equal to the given value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, float):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value != filter_value:
                return_list.append(row)
        return return_list

    # Note that the following six methods are identical to the ones above, except that the filter
    # values are obtained from `_get_date_row_pairs_to_filter`.
    def filter_date_greater(self,
                            row_expression_list: NestedList,
                            column_name: str,
                            value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows as an expression, a column, and a numerical value expression and
        returns all the rows where the value in that column is greater than the given value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_date_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, Date):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value > filter_value:
                return_list.append(row)
        return return_list

    def filter_date_greater_equals(self,
                                   row_expression_list: NestedList,
                                   column_name: str,
                                   value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows as an expression, a column, and a numerical value expression and
        returns all the rows where the value in that column is greater than or equal to the given
        value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_date_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, Date):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value >= filter_value:
                return_list.append(row)
        return return_list

    def filter_date_lesser(self,
                           row_expression_list: NestedList,
                           column_name: str,
                           value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows as an expression, a column, and a numerical value expression and
        returns all the rows where the value in that column is less than the given value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_date_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, Date):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value < filter_value:
                return_list.append(row)
        return return_list

    def filter_date_lesser_equals(self,
                                  row_expression_list: NestedList,
                                  column_name: str,
                                  value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows, a column, and a numerical value and returns all the rows where the
        value in that column is lesser than or equal to the given value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_date_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, Date):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value <= filter_value:
                return_list.append(row)
        return return_list

    def filter_date_equals(self,
                           row_expression_list: NestedList,
                           column_name: str,
                           value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows, a column, and a numerical value and returns all the rows where the
        value in that column equals the given value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_date_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, Date):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value == filter_value:
                return_list.append(row)
        return return_list

    def filter_date_not_equals(self,
                               row_expression_list: NestedList,
                               column_name: str,
                               value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows, a column, and a numerical value and returns all the rows where the
        value in that column is not equal to the given value.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        cell_row_pairs = self._get_date_row_pairs_to_filter(row_list, column_name)
        filter_value = self._handle_expression(value_expression)
        if not isinstance(filter_value, Date):
            raise ExecutionError(f"Invalid filter value: {value_expression}")
        return_list = []
        for cell_value, row in cell_row_pairs:
            if cell_value != filter_value:
                return_list.append(row)
        return return_list

    def filter_in(self,
                  row_expression_list: NestedList,
                  column_name: str,
                  value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows, a column, and a string value and returns all the rows where the value
        in that column contains the given string.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        expression_evaluation = self._handle_expression(value_expression)
        if isinstance(expression_evaluation, list):
            filter_value = expression_evaluation[0]
        elif isinstance(expression_evaluation, str):
            filter_value = expression_evaluation
        else:
            raise ExecutionError(f"Unexpected filter value for filter_in: {value_expression}")
        if not isinstance(filter_value, str):
            raise ExecutionError(f"Unexpected filter value for filter_in: {value_expression}")
        # Assuming filter value has underscores for spaces. The cell values also have underscores
        # for spaces, so we do not need to replace them here.
        result_list = []
        for row in row_list:
            if filter_value in row[column_name]:
                result_list.append(row)
        return result_list

    def filter_not_in(self,
                      row_expression_list: NestedList,
                      column_name: str,
                      value_expression: NestedList) -> List[Dict[str, str]]:
        """
        Takes a list of rows, a column, and a string value and returns all the rows where the value
        in that column does not contain the given string.
        """
        row_list = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        expression_evaluation = self._handle_expression(value_expression)
        if isinstance(expression_evaluation, list):
            filter_value = expression_evaluation[0]
        elif isinstance(expression_evaluation, str):
            filter_value = expression_evaluation
        else:
            raise ExecutionError(f"Unexpected filter value for filter_in: {value_expression}")
        if not isinstance(filter_value, str):
            raise ExecutionError(f"Unexpected filter value for filter_in: {value_expression}")
        # Assuming filter value has underscores for spaces. The cell values also have underscores
        # for spaces, so we do not need to replace them here.
        result_list = []
        for row in row_list:
            if filter_value not in row[column_name]:
                result_list.append(row)
        return result_list

    def first(self, row_expression_list: NestedList) -> List[Dict[str, str]]:
        """
        Takes an expression that evaluates to a list of rows, and returns the first one in that
        list.
        """
        row_list: List[Dict[str, str]] = self._handle_expression(row_expression_list)
        if not row_list:
            logger.warning("Trying to get first row from an empty list: %s", row_expression_list)
            return []
        return [row_list[0]]

    def last(self, row_expression_list: NestedList) -> List[Dict[str, str]]:
        """
        Takes an expression that evaluates to a list of rows, and returns the last one in that
        list.
        """
        row_list: List[Dict[str, str]] = self._handle_expression(row_expression_list)
        if not row_list:
            logger.warning("Trying to get last row from an empty list: %s", row_expression_list)
            return []
        return [row_list[-1]]

    def previous(self, row_expression_list: NestedList) -> List[Dict[str, str]]:
        """
        Takes an expression that evaluates to a single row, and returns the row (as a list to be
        consistent with the rest of the API), that occurs before the input row in the original set
        of rows. If the input row happens to be the top row, we will return an empty list.
        """
        row_list: List[Dict[str, str]] = self._handle_expression(row_expression_list)
        if not row_list:
            logger.warning("Trying to get the previous row from an empty list: %s",
                           row_expression_list)
            return []
        if len(row_list) > 1:
            logger.warning("Trying to get the previous row from a non-singleton list: %s",
                           row_expression_list)
        input_row_index = self._get_row_index(row_list[0])  # Take the first row.
        if input_row_index > 0:
            return [self.table_data[input_row_index - 1]]
        return []

    def next(self, row_expression_list: NestedList) -> List[Dict[str, str]]:
        """
        Takes an expression that evaluates to a single row, and returns the row (as a list to be
        consistent with the rest of the API), that occurs after the input row in the original set
        of rows. If the input row happens to be the last row, we will return an empty list.
        """
        row_list: List[Dict[str, str]] = self._handle_expression(row_expression_list)
        if not row_list:
            logger.warning("Trying to get the next row from an empty list: %s", row_expression_list)
            return []
        if len(row_list) > 1:
            logger.warning("Trying to get the next row from a non-singleton list: %s", row_expression_list)
        input_row_index = self._get_row_index(row_list[-1])  # Take the last row.
        if input_row_index < len(self.table_data) - 1 and input_row_index != -1:
            return [self.table_data[input_row_index + 1]]
        return []

    def count(self, row_expression_list: NestedList) -> float:
        """
        Takes an expression that evaluates to a a list of rows and returns their count (as a float
        to be consistent with the other functions like max that also return numbers).
        """
        row_list: List[Dict[str, str]] = self._handle_expression(row_expression_list)
        return float(len(row_list))

    def max(self,
            row_expression_list: NestedList,
            column_name: str) -> float:
        """
        Takes an expression list that evaluates to a  list of rows and a column name, and returns the max
        of the values under that column in those rows.
        """
        row_list: List[Dict[str, str]] = self._handle_expression(row_expression_list)
        cell_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)
        if not cell_row_pairs:
            return 0.0
        return max([value for value, _ in cell_row_pairs])

    def min(self,
            row_expression_list: NestedList,
            column_name: str) -> float:
        """
        Takes an expression list that evaluates to a  list of rows and a column, and returns the min
        of the values under that column in those rows.
        """
        row_list: List[Dict[str, str]] = self._handle_expression(row_expression_list)
        cell_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)
        if not cell_row_pairs:
            return 0.0
        return min([value for value, _ in cell_row_pairs])

    def sum(self,
            row_expression_list: NestedList,
            column_name) -> float:
        """
        Takes an expression list that evaluates to a  list of rows and a column, and returns the sum
        of the values under that column in those rows.
        """
        row_list: List[Dict[str, str]] = self._handle_expression(row_expression_list)
        cell_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)
        if not cell_row_pairs:
            return 0.0
        return sum([value for value, _ in cell_row_pairs])

    def average(self,
                row_expression_list: NestedList,
                column_name: str) -> float:
        """
        Takes an expression list that evaluates to a  list of rows and a column, and returns the mean
        of the values under that column in those rows.
        """
        row_list: List[Dict[str, str]] = self._handle_expression(row_expression_list)
        cell_row_pairs = self._get_number_row_pairs_to_filter(row_list, column_name)
        if not cell_row_pairs:
            return 0.0
        return sum([value for value, _ in cell_row_pairs]) / len(cell_row_pairs)

    def mode(self,
             row_expression_list: NestedList,
             column_name: str) -> List[str]:
        """
        Takes an expression that evaluates to a list of rows, and a column and returns the most
        frequent values (one or more) under that column in those rows.
        """
        row_list: List[Dict[str, str]] = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        value_frequencies: Dict[str, int] = defaultdict(int)
        max_frequency = 0
        most_frequent_list: List[str] = []
        for row in row_list:
            cell_value = row[column_name]
            value_frequencies[cell_value] += 1
            frequency = value_frequencies[cell_value]
            if frequency > max_frequency:
                max_frequency = frequency
                most_frequent_list = [cell_value]
            elif frequency == max_frequency:
                most_frequent_list.append(cell_value)
        return most_frequent_list

    def same_as(self,
                row_expression_list: NestedList,
                column_name: str) -> List[Dict[str, str]]:
        """
        Takes an expression that evaluates to a row, and a column and returns a list of rows from
        the full set of rows that contain the same value under the given column as the given row.
        """
        row_list: List[Dict[str, str]] = self._handle_expression(row_expression_list)
        if not row_list:
            return []
        if len(row_list) > 1:
            logger.warning("same_as function got multiple rows. Taking the first one: "
                           f"{row_expression_list}")
        cell_value = row_list[0][column_name]
        return_list = []
        for row in self.table_data:
            if row[column_name] == cell_value:
                return_list.append(row)
        return return_list

    def diff(self,
             first_row_expression_list: NestedList,
             second_row_expression_list: NestedList,
             column_name: str) -> float:
        """
        Takes an expressions that evaluate to two rows, and a column name, and returns the
        difference between the values under that column in those two rows.
        """
        first_row_list = self._handle_expression(first_row_expression_list)
        second_row_list = self._handle_expression(second_row_expression_list)
        if not first_row_list or not second_row_list:
            return 0.0
        if len(first_row_list) > 1:
            logger.warning("diff got multiple rows for first argument. Taking the first one: "
                           f"{first_row_expression_list}")
        if len(second_row_list) > 1:
            logger.warning("diff got multiple rows for second argument. Taking the first one: "
                           f"{second_row_expression_list}")
        first_row = first_row_list[0]
        second_row = second_row_list[0]
        try:
            first_value = float(first_row[column_name])
            second_value = float(second_row[column_name])
            return first_value - second_value
        except ValueError:
            raise ExecutionError(f"Invalid column for diff: {column_name}")

    @staticmethod
    def date(year_string: str, month_string: str, day_string: str) -> Date:
        """
        Takes three numbers as strings, and returns a ``Date`` object whose year, month, and day are
        the three numbers in that order.
        """
        try:
            year = int(str(year_string))
            month = int(str(month_string))
            day = int(str(day_string))
            return Date(year, month, day)
        except ValueError:
            raise ExecutionError(f"Invalid date! Got {year_string}, {month_string}, {day_string}")



from typing import List, Set
from overrides import overrides

from nltk.sem.logic import TRUTH_TYPE, BasicType, EntityType, Type

from allennlp.semparse.type_declarations.type_declaration import (ComplexType, HigherOrderType,
                                                                  NamedBasicType, NameMapper)


class NegateFilterType(HigherOrderType):
    """
    Because our negate filters are higher-order functions, we need to make an explicit class here,
    to make sure that we've overridden the right methods correctly.
    """
    def __init__(self, first, second):
        super().__init__(num_arguments=1, first=first, second=second)

    @overrides
    def substitute_any_type(self, basic_types: Set[BasicType]) -> List[Type]:
        # There's no ANY_TYPE in here, so we don't need to do any substitution.
        return [self]


# All constants default to ``EntityType`` in NLTK. For domains where constants of different types
# appear in the logical forms, we have a way of specifying ``constant_type_prefixes`` and passing
# them to the constructor of ``World``. However, in the NLVR language we defined, we see constants
# of just one type, number. So we let them default to ``EntityType``.
NUM_TYPE = EntityType()
BOX_TYPE = NamedBasicType("BOX")
OBJECT_TYPE = NamedBasicType("OBJECT")
COLOR_TYPE = NamedBasicType("COLOR")
SHAPE_TYPE = NamedBasicType("SHAPE")
OBJECT_FILTER_TYPE = ComplexType(OBJECT_TYPE, OBJECT_TYPE)
NEGATE_FILTER_TYPE = NegateFilterType(ComplexType(OBJECT_TYPE, OBJECT_TYPE),
                                      ComplexType(OBJECT_TYPE, OBJECT_TYPE))
BOX_MEMBERSHIP_TYPE = ComplexType(BOX_TYPE, OBJECT_TYPE)

BOX_COLOR_FILTER_TYPE = ComplexType(BOX_TYPE, ComplexType(COLOR_TYPE, BOX_TYPE))
BOX_SHAPE_FILTER_TYPE = ComplexType(BOX_TYPE, ComplexType(SHAPE_TYPE, BOX_TYPE))
BOX_COUNT_FILTER_TYPE = ComplexType(BOX_TYPE, ComplexType(NUM_TYPE, BOX_TYPE))
# This box filter returns boxes where a specified attribute is same or different
BOX_ATTRIBUTE_SAME_FILTER_TYPE = ComplexType(BOX_TYPE, BOX_TYPE)


ASSERT_COLOR_TYPE = ComplexType(OBJECT_TYPE, ComplexType(COLOR_TYPE, TRUTH_TYPE))
ASSERT_SHAPE_TYPE = ComplexType(OBJECT_TYPE, ComplexType(SHAPE_TYPE, TRUTH_TYPE))
ASSERT_BOX_COUNT_TYPE = ComplexType(BOX_TYPE, ComplexType(NUM_TYPE, TRUTH_TYPE))
ASSERT_OBJECT_COUNT_TYPE = ComplexType(OBJECT_TYPE, ComplexType(NUM_TYPE, TRUTH_TYPE))

BOX_EXISTS_TYPE = ComplexType(BOX_TYPE, TRUTH_TYPE)
OBJECT_EXISTS_TYPE = ComplexType(OBJECT_TYPE, TRUTH_TYPE)

BASIC_TYPES = {NUM_TYPE, BOX_TYPE, OBJECT_TYPE, COLOR_TYPE, SHAPE_TYPE}

name_mapper = NameMapper()  # pylint: disable=invalid-name

# Entities
name_mapper.map_name_with_signature("all_objects", OBJECT_TYPE)
name_mapper.map_name_with_signature("all_boxes", BOX_TYPE)
name_mapper.map_name_with_signature("color_black", COLOR_TYPE)
name_mapper.map_name_with_signature("color_blue", COLOR_TYPE)
name_mapper.map_name_with_signature("color_yellow", COLOR_TYPE)
name_mapper.map_name_with_signature("shape_triangle", SHAPE_TYPE)
name_mapper.map_name_with_signature("shape_square", SHAPE_TYPE)
name_mapper.map_name_with_signature("shape_circle", SHAPE_TYPE)


# Attribute function
name_mapper.map_name_with_signature("object_in_box", BOX_MEMBERSHIP_TYPE)


# Assert functions
name_mapper.map_name_with_signature("object_color_all_equals", ASSERT_COLOR_TYPE)
name_mapper.map_name_with_signature("object_color_any_equals", ASSERT_COLOR_TYPE)
name_mapper.map_name_with_signature("object_color_none_equals", ASSERT_COLOR_TYPE)
name_mapper.map_name_with_signature("object_shape_all_equals", ASSERT_SHAPE_TYPE)
name_mapper.map_name_with_signature("object_shape_any_equals", ASSERT_SHAPE_TYPE)
name_mapper.map_name_with_signature("object_shape_none_equals", ASSERT_SHAPE_TYPE)
name_mapper.map_name_with_signature("box_count_equals", ASSERT_BOX_COUNT_TYPE)
name_mapper.map_name_with_signature("box_count_not_equals", ASSERT_BOX_COUNT_TYPE)
name_mapper.map_name_with_signature("box_count_greater", ASSERT_BOX_COUNT_TYPE)
name_mapper.map_name_with_signature("box_count_greater_equals", ASSERT_BOX_COUNT_TYPE)
name_mapper.map_name_with_signature("box_count_lesser", ASSERT_BOX_COUNT_TYPE)
name_mapper.map_name_with_signature("box_count_lesser_equals", ASSERT_BOX_COUNT_TYPE)
name_mapper.map_name_with_signature("object_count_equals", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_count_not_equals", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_count_greater", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_count_greater_equals", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_count_lesser", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_count_lesser_equals", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_color_count_equals", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_color_count_not_equals", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_color_count_greater", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_color_count_greater_equals", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_color_count_lesser", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_color_count_lesser_equals", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_shape_count_equals", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_shape_count_not_equals", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_shape_count_greater", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_shape_count_greater_equals", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_shape_count_lesser", ASSERT_OBJECT_COUNT_TYPE)
name_mapper.map_name_with_signature("object_shape_count_lesser_equals", ASSERT_OBJECT_COUNT_TYPE)

name_mapper.map_name_with_signature("box_exists", BOX_EXISTS_TYPE)
name_mapper.map_name_with_signature("object_exists", OBJECT_EXISTS_TYPE)


# Box filter functions
name_mapper.map_name_with_signature("member_count_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_count_not_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_shape_all_equals", BOX_SHAPE_FILTER_TYPE)
name_mapper.map_name_with_signature("member_shape_any_equals", BOX_SHAPE_FILTER_TYPE)
name_mapper.map_name_with_signature("member_shape_none_equals", BOX_SHAPE_FILTER_TYPE)
name_mapper.map_name_with_signature("member_color_all_equals", BOX_COLOR_FILTER_TYPE)
name_mapper.map_name_with_signature("member_color_any_equals", BOX_COLOR_FILTER_TYPE)
name_mapper.map_name_with_signature("member_color_none_equals", BOX_COLOR_FILTER_TYPE)
name_mapper.map_name_with_signature("member_count_greater", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_count_greater_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_count_lesser", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_count_lesser_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_color_count_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_color_count_not_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_color_count_greater", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_color_count_greater_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_color_count_lesser", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_color_count_lesser_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_shape_count_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_shape_count_not_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_shape_count_greater", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_shape_count_greater_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_shape_count_lesser", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_shape_count_lesser_equals", BOX_COUNT_FILTER_TYPE)
name_mapper.map_name_with_signature("member_shape_same", BOX_ATTRIBUTE_SAME_FILTER_TYPE)
name_mapper.map_name_with_signature("member_color_same", BOX_ATTRIBUTE_SAME_FILTER_TYPE)
name_mapper.map_name_with_signature("member_shape_different", BOX_ATTRIBUTE_SAME_FILTER_TYPE)
name_mapper.map_name_with_signature("member_color_different", BOX_ATTRIBUTE_SAME_FILTER_TYPE)


# Object filter functions
name_mapper.map_name_with_signature("black", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("blue", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("yellow", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("same_color", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("triangle", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("square", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("circle", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("same_shape", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("touch_wall", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("touch_corner", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("touch_top", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("touch_bottom", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("touch_left", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("touch_right", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("touch_object", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("above", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("below", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("top", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("bottom", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("small", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("medium", OBJECT_FILTER_TYPE)
name_mapper.map_name_with_signature("big", OBJECT_FILTER_TYPE)

name_mapper.map_name_with_signature("negate_filter", NEGATE_FILTER_TYPE)

# Adding numbers because they commonly occur in utterances. They're usually between 1 and 9. Since
# there are not too many of these productions, we're adding them to the global mapping instead of a
# local mapping in each world.
for num in range(1, 10):
    num_string = str(num)
    name_mapper.map_name_with_signature(name=num_string, signature=NUM_TYPE, alias=num_string)


COMMON_NAME_MAPPING = name_mapper.common_name_mapping
COMMON_TYPE_SIGNATURE = name_mapper.common_type_signature



from allennlp.common.util import JsonDict


class Object:
    """
    ``Objects`` are the geometric shapes in the NLVR domain. They have values for attributes shape,
    color, x_loc, y_loc and size. We take a dict read from the JSON file and store it here, and
    define a get method for getting the attribute values. We need this to be hashable because need
    to make sets of ``Objects`` during execution, which get passed around between functions.

    Parameters
    ----------
    attributes : ``JsonDict``
        The dict for each object from the json file.
    """
    def __init__(self, attributes: JsonDict, box_id: str) -> None:
        object_color = attributes["color"].lower()
        # The dataset has a hex code only for blue for some reason.
        if object_color.startswith("#"):
            self.color = "blue"
        else:
            self.color = object_color
        object_shape = attributes["type"].lower()
        self.shape = object_shape
        self.x_loc = attributes["x_loc"]
        self.y_loc = attributes["y_loc"]
        self.size = attributes["size"]
        self._box_id = box_id

    def __str__(self):
        if self.size == 10:
            size = "small"
        elif self.size == 20:
            size = "medium"
        else:
            size = "big"
        return f"{size} {self.color} {self.shape} at ({self.x_loc}, {self.y_loc}) in {self._box_id}"

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)



import os
import pytest

try:
    import nbformat
    from nbconvert.preprocessors.execute import CellExecutionError
    from nbconvert.preprocessors import ExecutePreprocessor
except ModuleNotFoundError:
    print("jupyter must be installed in order to run notebook tests. "
          "To install with pip, run: pip install jupyter")

from allennlp.common.testing import AllenNlpTestCase

class TestNotebooks(AllenNlpTestCase):
    @pytest.mark.skip(reason="these tests are brittle and fail with small dependency changes")
    def test_vocabulary_tutorial(self):
        assert self.execute_notebook("tutorials/notebooks/vocabulary.ipynb")

    @pytest.mark.skip(reason="these tests are brittle and fail with small dependency changes")
    def test_data_pipeline_tutorial(self):
        assert self.execute_notebook("tutorials/notebooks/data_pipeline.ipynb")

    @pytest.mark.skip(reason="these tests are brittle and fail with small dependency changes")
    def test_embedding_tokens_tutorial(self):
        assert self.execute_notebook("tutorials/notebooks/embedding_tokens.ipynb")

    @staticmethod
    def execute_notebook(notebook_path: str):
        with open(notebook_path, encoding='utf-8') as notebook:
            contents = nbformat.read(notebook, as_version=4)

        execution_processor = ExecutePreprocessor(timeout=60, kernel_name="python3")
        try:
            # Actually execute the notebook in the current working directory.
            execution_processor.preprocess(contents, {'metadata': {'path': os.getcwd()}})
            return True
        except CellExecutionError:
            # This is a big chunk of JSON, but the stack trace makes it reasonably
            # clear which cell the error occurred in, so fixing it by actually
            # running the notebook will probably be easier.
            print(contents)
            return False



"""
AllenNLP uses most
`PyTorch learning rate schedulers <http://pytorch.org/docs/master/optim.html#how-to-adjust-learning-rate>`_,
with a thin wrapper to allow registering them and instantiating them ``from_params``.

The available learning rate schedulers from PyTorch are

* `"step" <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.StepLR>`_
* `"multi_step" <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.MultiStepLR>`_
* `"exponential" <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ExponentialLR>`_
* `"reduce_on_plateau" <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ReduceLROnPlateau>`_

In addition, AllenNLP also provides a Noam schedule and `cosine with restarts
<https://arxiv.org/abs/1608.03983>`_, which are registered as "noam" and "cosine", respectively.
"""

# During training using the AllenNLP `Trainer`, this is the API and calling
#sequence for `step` and `step_batch`.  Also note that `step` is called
#once in `torch.optim.lr_scheduler._LRScheduler.__init__`.
#
#   scheduler = ... # creates scheduler, calls self.step(last_epoch + 1) in __init__
#
#   batch_num_total = 0
#   for epoch in range(num_epochs):
#       for batch in batchs_in_epoch:
#           # compute loss, update parameters with current learning rates
#           # call step_batch AFTER updating parameters
#           batch_num_total += 1
#           scheduler.step_batch(batch_num_total)
#       # call step() at the END of each epoch
#       scheduler.step(validation_metrics, epoch)

import logging
from typing import Optional, List

import numpy as np
import torch.optim.lr_scheduler
from overrides import overrides

from allennlp.common.checks import ConfigurationError
from allennlp.common.params import Params
from allennlp.common.registrable import Registrable


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class LearningRateScheduler(Registrable):
    """
    This class just allows us to implement ``Registrable`` for Pytorch :class:`LRSchedulers`.
    """
    def __init__(self, lr_scheduler) -> None:
        self.lr_scheduler = lr_scheduler

    def step(self, metric: float, epoch: Optional[int] = None):
        raise NotImplementedError

    def step_batch(self, batch_num_total: Optional[int]):
        if batch_num_total is not None:
            if hasattr(self.lr_scheduler, 'step_batch'):
                self.lr_scheduler.step_batch(batch_num_total)
            return

    # Requires custom from_params
    @classmethod
    def from_params(cls, optimizer: torch.optim.Optimizer, params: Params):  # type: ignore
        # pylint: disable=arguments-differ
        scheduler = params.pop_choice("type", LearningRateScheduler.list_available())

        schedulers = LearningRateScheduler.by_name(scheduler)(optimizer, **params.as_dict())  # type: ignore
        if isinstance(schedulers, torch.optim.lr_scheduler.ReduceLROnPlateau):
            return LearningRateWithMetricsWrapper(schedulers)
        else:
            return LearningRateWithoutMetricsWrapper(schedulers)


class LearningRateWithoutMetricsWrapper(LearningRateScheduler):
    """
    A wrapper around learning rate schedulers that do not require metrics
    """
    def __init__(self, lr_scheduler: torch.optim.lr_scheduler._LRScheduler) -> None:  # pylint: disable=protected-access
        super().__init__(lr_scheduler)
        self.lr_scheduler = lr_scheduler

    @overrides
    def step(self, metric: float, epoch: Optional[int] = None):
        self.lr_scheduler.step(epoch)


class LearningRateWithMetricsWrapper(LearningRateScheduler):
    """
    A wrapper around learning rate schedulers that require metrics,
    At the moment there is only a single instance of this lrs. It is the ReduceLROnPlateau
    """
    def __init__(self, lr_scheduler: torch.optim.lr_scheduler.ReduceLROnPlateau) -> None:
        super().__init__(lr_scheduler)
        self.lr_scheduler = lr_scheduler

    @overrides
    def step(self, metric: float, epoch: Optional[int] = None):
        if metric is None:
            raise ConfigurationError("The reduce_on_plateau learning rate scheduler requires "
                                     "a validation metric to compute the schedule and therefore "
                                     "must be used with a validation dataset.")
        self.lr_scheduler.step(metric, epoch)


class NoamLR(torch.optim.lr_scheduler._LRScheduler): # pylint: disable=protected-access
    """
    Implements the Noam Learning rate schedule. This corresponds to increasing the learning rate
    linearly for the first ``warmup_steps`` training steps, and decreasing it thereafter proportionally
    to the inverse square root of the step number, scaled by the inverse square root of the
    dimensionality of the model. Time will tell if this is just madness or it's actually important.

    Parameters
    ----------
    model_size : ``int``, required.
        The hidden size parameter which dominates the number of parameters in your model.
    warmup_steps: ``int``, required.
        The number of steps to linearly increase the learning rate.
    factor : ``float``, optional (default = 1.0).
        The overall scale factor for the learning rate decay.
    """
    def __init__(self,
                 optimizer: torch.optim.Optimizer,
                 model_size: int,
                 warmup_steps: int,
                 factor: float = 1.0,
                 last_epoch: int = -1) -> None:
        self.warmup_steps = warmup_steps
        self.factor = factor
        self.model_size = model_size
        super().__init__(optimizer, last_epoch=last_epoch)

    def step(self, epoch=None):
        pass

    def step_batch(self, epoch=None):
        if epoch is None:
            epoch = self.last_epoch + 1
        self.last_epoch = epoch
        for param_group, learning_rate in zip(self.optimizer.param_groups, self.get_lr()):
            param_group['lr'] = learning_rate

    def get_lr(self):
        step = max(self.last_epoch, 1)
        scale = self.factor *  (self.model_size ** (-0.5) *
                                min(step ** (-0.5), step * self.warmup_steps ** (-1.5)))

        return [scale for _ in range(len(self.base_lrs))]


class SlantedTriangular(torch.optim.lr_scheduler._LRScheduler): # pylint: disable=protected-access
    """
    Implements the Slanted Triangular Learning Rate schedule with optional gradual
    unfreezing. The schedule corresponds to first linearly increasing the learning
    rate and annealing the learning based on a fixed ratio.

    If we gradually unfreeze, then in the first epoch of training, only the top
    layer is trained; in the second epoch, the top two layers are trained, etc.
    During freezing, the learning rate is increased and annealed over one epoch.
    After freezing finished, the learning rate is increased and annealed over
    the remaining training iterations.

    Note that with this schedule, early stopping should typically be avoided.

    Parameters
    ----------
    num_epochs : ``int``, required.
        The total number of epochs for which the model should be trained.
    num_steps_per_epoch: ``int``, required.
        The number of steps (updates, batches) per training epoch.
    cut_frac: ``float``, optional (default = 0.1).
        The fraction of the steps to increase the learning rate.
    ratio: ``float``, optional (default = 32).
        The ratio of the smallest to the (largest) base learning rate.
    gradual_unfreezing: ``bool``, optional (default = False).
        Whether gradual unfreezing should be used.
    discriminative_fine_tuning: ``bool``, optional (default = False).
        Whether discriminative fine-tuning (different learning rates per layer)
        are used.
    decay_factor: ``float``, optional (default = 0.38).
        The decay factor by which the learning rate is reduced with
        discriminative fine-tuning when going a layer deeper.
    """
    def __init__(self,
                 optimizer: torch.optim.Optimizer,
                 num_epochs: int,
                 num_steps_per_epoch: int,
                 cut_frac: float = 0.1,
                 ratio: int = 32,
                 last_epoch: int = -1,
                 gradual_unfreezing: bool = False,
                 discriminative_fine_tuning: bool = False,
                 decay_factor: float = 0.38) -> None:
        self.num_epochs = num_epochs
        self.num_steps_per_epoch = num_steps_per_epoch
        self.cut_frac = cut_frac
        self.ratio = ratio
        self.gradual_unfreezing = gradual_unfreezing
        self.freezing_current = self.gradual_unfreezing
        self.is_first_epoch = True
        # track the actual number of steps for each epoch
        self.batch_num_total_epoch_end: List[int] = []
        if self.gradual_unfreezing:
            assert not optimizer.param_groups[-1]["params"], \
                "The default group should be empty."
        if self.gradual_unfreezing or discriminative_fine_tuning:
            assert len(optimizer.param_groups) > 2, \
                "There should be at least 3 param_groups (2 + empty default group)" \
                " for gradual unfreezing / discriminative fine-tuning to make sense."
        super().__init__(optimizer, last_epoch=last_epoch)
        if discriminative_fine_tuning:
            # skip the last param_group if it is has no parameters
            exponent = 0
            for i in range(len(self.base_lrs)-1, -1, -1):
                param_group = optimizer.param_groups[i]
                if param_group['params']:
                    param_group['lr'] = self.base_lrs[i] * decay_factor ** exponent
                    self.base_lrs[i] = param_group['lr']
                    exponent += 1
        # set up for the first batch
        self.last_batch_num_total = -1
        self.step_batch(0)

    def step(self, epoch=None):
        if len(self.batch_num_total_epoch_end) == 0: # pylint: disable=len-as-condition
            self.batch_num_total_epoch_end.append(0)
        else:
            self.batch_num_total_epoch_end.append(self.last_batch_num_total)

        if self.gradual_unfreezing:
            # the method is called once when initialising before the
            # first epoch (epoch 0) and then always at the end of each
            # epoch; so the first time, with epoch id 0, we want to set
            # up for epoch #1; the second time, still with epoch id 0,
            # we want to set up for epoch #2, etc.
            num_layers_to_unfreeze = epoch + 1 if self.is_first_epoch else epoch + 2
            if self.is_first_epoch:
                self.is_first_epoch = False
            if num_layers_to_unfreeze >= len(self.optimizer.param_groups)-1:
                logger.info('Gradual unfreezing finished. Training all layers.')
                self.freezing_current = False
            else:
                logger.info(f'Gradual unfreezing. Training only the top {num_layers_to_unfreeze} layers.')
            for i, param_group in enumerate(reversed(self.optimizer.param_groups)):
                for param in param_group["params"]:
                    # i = 0 is the default group; we care about i > 0
                    param.requires_grad = bool(i <= num_layers_to_unfreeze)

    def step_batch(self, batch_num_total=None):
        if batch_num_total is None:
            batch_num_total = self.last_batch_num_total + 1
        self.last_batch_num_total = batch_num_total
        for param_group, learning_rate in zip(self.optimizer.param_groups, self.get_lr()):
            param_group['lr'] = learning_rate

    def get_lr(self):
        # get the actual number of batches per epoch seen in training
        if len(self.batch_num_total_epoch_end) > 1:
            # have finished an epoch
            actual_num_steps_per_epoch = int(
                    self.batch_num_total_epoch_end[-1] /
                    (len(self.batch_num_total_epoch_end) - 1)
            )
        else:
            actual_num_steps_per_epoch = max(self.num_steps_per_epoch,
                                             self.last_batch_num_total)

        if self.freezing_current:
            # if we still freeze, we restrict the schedule to the current epoch
            num_steps = actual_num_steps_per_epoch
            step = min(self.last_batch_num_total - self.batch_num_total_epoch_end[-1],
                       num_steps)
        else:
            # otherwise we use the schedule for the rest of training
            if not self.gradual_unfreezing:
                frozen_steps = 0
            else:
                num_frozen_epochs = len(self.optimizer.param_groups) - 2
                frozen_steps = self.batch_num_total_epoch_end[num_frozen_epochs]
            num_steps = self.num_epochs * actual_num_steps_per_epoch - frozen_steps
            step = min(self.last_batch_num_total - frozen_steps,
                       num_steps)
        cut = int(num_steps * self.cut_frac)
        prop = step / cut if step < cut else 1 - (step - cut) / (num_steps - cut)
        return [lr * (1 + prop * (self.ratio - 1)) / self.ratio for lr in self.base_lrs]


class CosineWithRestarts(torch.optim.lr_scheduler._LRScheduler):  # pylint: disable=protected-access
    """
    Cosine annealing with restarts.

    This is described in the paper https://arxiv.org/abs/1608.03983.

    Parameters
    ----------
    optimizer : ``torch.optim.Optimizer``

    t_initial : ``int``
        The number of iterations within the first cycle.

    t_mul : ``float``, optional (default=1)
        Determines the number of iterations in the i-th decay cycle, which is the
        length of the last cycle multiplied by ``t_mul``.

    eta_min : ``float``, optional (default=0)
        The minimum learning rate.

    eta_mul : ``float``, optional (default=1)
        Determines the initial learning rate for the i-th decay cycle, which is the
        last initial learning rate multiplied by ``m_mul``.

    last_epoch : ``int``, optional (default=-1)
        The index of the last epoch. This is used when restarting.

    """

    def __init__(self,
                 optimizer: torch.optim.Optimizer,
                 t_initial: int,
                 t_mul: float = 1.,
                 eta_min: float = 0.,
                 eta_mul: float = 1.,
                 last_epoch: int = -1) -> None:
        assert t_initial > 0
        assert eta_min >= 0
        if t_initial == 1 and t_mul == 1 and eta_mul == 1:
            logger.warning("Cosine annealing scheduler will have no effect on the learning "
                           "rate since t_initial = t_mul = eta_mul = 1.")
        self.t_initial = t_initial
        self.t_mul = t_mul
        self.eta_min = eta_min
        self.eta_mul = eta_mul
        self._last_restart: int = 0
        self._cycle_counter: int = 0
        self._cycle_len: int = t_initial
        self._n_restarts: int = 0
        self._initialized: bool = False
        super(CosineWithRestarts, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        """Get updated learning rate."""
        # HACK: We need to check if this is the first time ``self.get_lr()`` was called,
        # since ``torch.optim.lr_scheduler._LRScheduler`` will call ``self.get_lr()``
        # when first initialized, but the learning rate should remain unchanged
        # for the first epoch.
        if not self._initialized:
            self._initialized = True
            return self.base_lrs

        step = self.last_epoch + 1
        self._cycle_counter = step - self._last_restart

        if self._cycle_counter % self._cycle_len == 0:
            self._n_restarts += 1
            self._cycle_counter = 0
            self._last_restart = step

        base_lrs = [lr * self.eta_mul**self._n_restarts for lr in self.base_lrs]
        self._cycle_len = int(self.t_initial * self.t_mul**self._n_restarts)

        lrs = [
                self.eta_min + ((lr - self.eta_min) / 2) * (
                        np.cos(np.pi * (self._cycle_counter % self._cycle_len) / self._cycle_len) + 1
                )
                for lr in base_lrs
        ]

        return lrs


# We just use the Pytorch LRSchedulers, so here we force them into
# Registry._registry so we can build them from params.
Registrable._registry[LearningRateScheduler] = {   # pylint: disable=protected-access
        "step": torch.optim.lr_scheduler.StepLR,
        "multi_step": torch.optim.lr_scheduler.MultiStepLR,
        "exponential": torch.optim.lr_scheduler.ExponentialLR,
        "reduce_on_plateau": torch.optim.lr_scheduler.ReduceLROnPlateau,
        "cosine": CosineWithRestarts,
        "noam": NoamLR,
        "slanted_triangular": SlantedTriangular,
}



# pylint: disable=no-self-use,invalid-name,too-many-public-methods
from typing import List
from functools import partial

from allennlp.common.testing import AllenNlpTestCase
from allennlp.data.tokenizers import Token
from allennlp.semparse.contexts import TableQuestionContext
from allennlp.semparse.type_declarations import wikitables_variable_free as types
from allennlp.semparse.worlds import WikiTablesVariableFreeWorld
from allennlp.semparse import ParsingError


def check_productions_match(actual_rules: List[str], expected_right_sides: List[str]):
    actual_right_sides = [rule.split(' -> ')[1] for rule in actual_rules]
    assert set(actual_right_sides) == set(expected_right_sides)


class TestWikiTablesVariableFreeWorld(AllenNlpTestCase):
    def setUp(self):
        super().setUp()
        question_tokens = [Token(x) for x in ['what', 'was', 'the', 'last', 'year', '2013', '?']]
        self.table_file = self.FIXTURES_ROOT / 'data' / 'wikitables' / 'sample_table.tagged'
        self.table_context = TableQuestionContext.read_from_file(self.table_file, question_tokens)
        self.world_with_2013 = WikiTablesVariableFreeWorld(self.table_context)
        usl_league_tokens = [Token(x) for x in ['what', 'was', 'the', 'last', 'year', 'with', 'usl',
                                                'a', 'league', '?']]
        self.world_with_usl_a_league = self._get_world_with_question_tokens(usl_league_tokens)

    def _get_world_with_question_tokens(self, tokens: List[Token]) -> WikiTablesVariableFreeWorld:
        table_context = TableQuestionContext.read_from_file(self.table_file, tokens)
        world = WikiTablesVariableFreeWorld(table_context)
        return world

    def test_get_valid_actions_returns_correct_set(self):
        # This test is long, but worth it.  These are all of the valid actions in the grammar, and
        # we want to be sure they are what we expect.
        valid_actions = self.world_with_2013.get_valid_actions()
        assert set(valid_actions.keys()) == {
                "<r,<g,s>>",
                "<r,<f,<n,r>>>",
                "<r,<c,r>>",
                "<r,<g,r>>",
                "<r,<r,<f,n>>>",
                "<r,<t,<s,r>>>",
                "<n,<n,<n,d>>>",
                "<r,<m,<d,r>>>",
                "<r,<f,n>>",
                "<r,r>",
                "<r,n>",
                "d",
                "n",
                "s",
                "m",
                "t",
                "f",
                "r",
                "@start@",
                }

        check_productions_match(valid_actions['<r,<g,s>>'],
                                ['mode', 'select'])

        check_productions_match(valid_actions['<r,<f,<n,r>>>'],
                                ['filter_number_equals', 'filter_number_greater',
                                 'filter_number_greater_equals', 'filter_number_lesser',
                                 'filter_number_lesser_equals', 'filter_number_not_equals'])

        check_productions_match(valid_actions['<r,<c,r>>'],
                                ['argmax', 'argmin'])

        check_productions_match(valid_actions['<r,<g,r>>'],
                                ['same_as'])

        check_productions_match(valid_actions['<r,<r,<f,n>>>'],
                                ['diff'])

        check_productions_match(valid_actions['<r,<t,<s,r>>>'],
                                ['filter_in', 'filter_not_in'])

        check_productions_match(valid_actions['<n,<n,<n,d>>>'],
                                ['date'])

        check_productions_match(valid_actions['<r,<m,<d,r>>>'],
                                ['filter_date_equals', 'filter_date_greater',
                                 'filter_date_greater_equals', 'filter_date_lesser',
                                 'filter_date_lesser_equals', 'filter_date_not_equals'])

        check_productions_match(valid_actions['<r,<f,n>>'],
                                ['average', 'max', 'min', 'sum'])

        check_productions_match(valid_actions['<r,r>'],
                                ['first', 'last', 'next', 'previous'])

        check_productions_match(valid_actions['<r,n>'],
                                ['count'])

        # These are the columns in table, and are instance specific.
        check_productions_match(valid_actions['m'],
                                ['date_column:year'])

        check_productions_match(valid_actions['f'],
                                ['number_column:avg_attendance',
                                 'number_column:division'])

        check_productions_match(valid_actions['t'],
                                ['string_column:league',
                                 'string_column:playoffs',
                                 'string_column:open_cup',
                                 'string_column:regular_season'])

        check_productions_match(valid_actions['@start@'],
                                ['d', 'n', 's'])

        # The question does not produce any strings. It produces just a number.
        check_productions_match(valid_actions['s'],
                                ['[<r,<g,s>>, r, m]',
                                 '[<r,<g,s>>, r, f]',
                                 '[<r,<g,s>>, r, t]'])

        check_productions_match(valid_actions['d'],
                                ['[<n,<n,<n,d>>>, n, n, n]'])

        check_productions_match(valid_actions['n'],
                                ['2013',
                                 '-1',
                                 '[<r,<f,n>>, r, f]',
                                 '[<r,<r,<f,n>>>, r, r, f]',
                                 '[<r,n>, r]'])

        check_productions_match(valid_actions['r'],
                                ['all_rows',
                                 '[<r,<m,<d,r>>>, r, m, d]',
                                 '[<r,<g,r>>, r, m]',
                                 '[<r,<g,r>>, r, f]',
                                 '[<r,<g,r>>, r, t]',
                                 '[<r,<c,r>>, r, m]',
                                 '[<r,<c,r>>, r, f]',
                                 '[<r,<f,<n,r>>>, r, f, n]',
                                 '[<r,<t,<s,r>>>, r, t, s]',
                                 '[<r,r>, r]'])

    def test_get_valid_actions_in_world_without_number_columns(self):
        question_tokens = [Token(x) for x in ['what', 'was', 'the', 'first', 'title', '?']]
        table_file = self.FIXTURES_ROOT / 'data' / 'corenlp_processed_tables' / 'TEST-6.table'
        table_context = TableQuestionContext.read_from_file(table_file, question_tokens)
        # The table does not have a number column.
        assert "number" not in table_context.column_types.values()
        world = WikiTablesVariableFreeWorld(table_context)
        actions = world.get_valid_actions()
        assert set(actions.keys()) == {
                "<r,<g,s>>",
                "<r,<c,r>>",
                "<r,<g,r>>",
                "<r,<t,<s,r>>>",
                "<n,<n,<n,d>>>",
                "<r,<m,<d,r>>>",
                "<r,r>",
                "<r,n>",
                "d",
                "n",
                "s",
                "m",
                "t",
                "r",
                "@start@",
                }
        assert set([str(type_) for type_ in world.get_basic_types()]) == {'n', 'd', 's', 'r', 't',
                                                                          'm', 'g', 'c'}
        check_productions_match(actions['s'],
                                ['[<r,<g,s>>, r, m]',
                                 '[<r,<g,s>>, r, t]'])

    def test_get_valid_actions_in_world_without_date_columns(self):
        question_tokens = [Token(x) for x in ['what', 'was', 'the', 'first', 'title', '?']]
        table_file = self.FIXTURES_ROOT / 'data' / 'corenlp_processed_tables' / 'TEST-4.table'
        table_context = TableQuestionContext.read_from_file(table_file, question_tokens)
        # The table does not have a date column.
        assert "date" not in table_context.column_types.values()
        world = WikiTablesVariableFreeWorld(table_context)
        actions = world.get_valid_actions()
        assert set(actions.keys()) == {
                "<r,<g,s>>",
                "<r,<f,<n,r>>>",
                "<r,<c,r>>",
                "<r,<g,r>>",
                "<r,<r,<f,n>>>",
                "<r,<t,<s,r>>>",
                "<n,<n,<n,d>>>",
                "<r,<f,n>>",
                "<r,r>",
                "<r,n>",
                "d",
                "n",
                "s",
                "t",
                "f",
                "r",
                "@start@",
                }
        assert set([str(type_) for type_ in world.get_basic_types()]) == {'n', 'd', 's', 'r', 't',
                                                                          'f', 'g', 'c'}
        check_productions_match(actions['s'],
                                ['[<r,<g,s>>, r, f]',
                                 '[<r,<g,s>>, r, t]'])

    def test_get_valid_actions_in_world_without_comparable_columns(self):
        question_tokens = [Token(x) for x in ['what', 'was', 'the', 'first', 'title', '?']]
        table_file = self.FIXTURES_ROOT / 'data' / 'corenlp_processed_tables' / 'TEST-1.table'
        table_context = TableQuestionContext.read_from_file(table_file, question_tokens)
        # The table does not have date or number columns.
        assert "date" not in table_context.column_types.values()
        assert "number" not in table_context.column_types.values()
        world = WikiTablesVariableFreeWorld(table_context)
        actions = world.get_valid_actions()
        assert set(actions.keys()) == {
                "<r,<g,s>>",
                "<r,<g,r>>",
                "<r,<t,<s,r>>>",
                "<n,<n,<n,d>>>",
                "<r,r>",
                "<r,n>",
                "d",
                "n",
                "s",
                "t",
                "r",
                "@start@",
                }
        assert set([str(type_) for type_ in world.get_basic_types()]) == {'n', 'd', 's', 'r', 't', 'g'}

    def test_parsing_logical_form_with_string_not_in_question_fails(self):
        logical_form_with_usl_a_league = """(select (filter_in all_rows string_column:league usl_a_league)
                                             date_column:year)"""
        logical_form_with_2013 = """(select (filter_date_greater all_rows date_column:year (date 2013 -1 -1))
                                     date_column:year)"""
        with self.assertRaises(ParsingError):
            self.world_with_2013.parse_logical_form(logical_form_with_usl_a_league)
            self.world_with_usl_a_league.parse_logical_form(logical_form_with_2013)

    @staticmethod
    def _get_alias(types_, name) -> str:
        if name in types_.generic_name_mapper.common_name_mapping:
            return types_.generic_name_mapper.get_alias(name)
        elif name in types_.string_column_name_mapper.common_name_mapping:
            return types_.string_column_name_mapper.get_alias(name)
        elif name in types_.number_column_name_mapper.common_name_mapping:
            return types_.number_column_name_mapper.get_alias(name)
        elif name in types_.date_column_name_mapper.common_name_mapping:
            return types_.date_column_name_mapper.get_alias(name)
        else:
            return types_.comparable_column_name_mapper.get_alias(name)

    def test_world_processes_logical_forms_correctly(self):
        logical_form = "(select (filter_in all_rows string_column:league string:usl_a_league) date_column:year)"
        expression = self.world_with_usl_a_league.parse_logical_form(logical_form)
        f = partial(self._get_alias, types)
        # Cells (and parts) get mapped to strings.
        # Column names are mapped in local name mapping. For the global names, we can get their
        # aliases from the name mapper.
        assert str(expression) == f"{f('select')}({f('filter_in')}({f('all_rows')},C2,string:usl_a_league),C0)"

    def test_world_gets_correct_actions(self):
        logical_form = "(select (filter_in all_rows string_column:league string:usl_a_league) date_column:year)"
        expression = self.world_with_usl_a_league.parse_logical_form(logical_form)
        expected_sequence = ['@start@ -> s', 's -> [<r,<g,s>>, r, m]', '<r,<g,s>> -> select',
                             'r -> [<r,<t,<s,r>>>, r, t, s]', '<r,<t,<s,r>>> -> filter_in',
                             'r -> all_rows', 't -> string_column:league', 's -> string:usl_a_league',
                             'm -> date_column:year']
        assert self.world_with_usl_a_league.get_action_sequence(expression) == expected_sequence

    def test_world_gets_logical_form_from_actions(self):
        logical_form = "(select (filter_in all_rows string_column:league string:usl_a_league) date_column:year)"
        expression = self.world_with_usl_a_league.parse_logical_form(logical_form)
        action_sequence = self.world_with_usl_a_league.get_action_sequence(expression)
        reconstructed_logical_form = self.world_with_usl_a_league.get_logical_form(action_sequence)
        assert logical_form == reconstructed_logical_form

    def test_world_processes_logical_forms_with_number_correctly(self):
        tokens = [Token(x) for x in ['when', 'was', 'the', 'attendance', 'higher', 'than', '3000',
                                     '?']]
        world = self._get_world_with_question_tokens(tokens)
        logical_form = """(select (filter_number_greater all_rows number_column:avg_attendance 3000)
                           date_column:year)"""
        expression = world.parse_logical_form(logical_form)
        f = partial(self._get_alias, types)
        # Cells (and parts) get mapped to strings.
        # Column names are mapped in local name mapping. For the global names, we can get their
        # aliases from the name mapper.
        assert str(expression) == f"{f('select')}({f('filter_number_greater')}({f('all_rows')},C6,num:3000),C0)"

    def test_world_processes_logical_forms_with_date_correctly(self):
        logical_form = """(select (filter_date_greater all_rows date_column:year (date 2013 -1 -1))
                           date_column:year)"""
        expression = self.world_with_2013.parse_logical_form(logical_form)
        f = partial(self._get_alias, types)
        # Cells (and parts) get mapped to strings.
        # Column names are mapped in local name mapping. For the global names, we can get their
        # aliases from the name mapper.
        assert str(expression) == \
        f"{f('select')}({f('filter_date_greater')}({f('all_rows')},C0,{f('date')}(num:2013,num:~1,num:~1)),C0)"

    def test_get_agenda(self):
        tokens = [Token(x) for x in ['what', 'was', 'the', 'last', 'year', '2000', '?']]
        world = self._get_world_with_question_tokens(tokens)
        assert set(world.get_agenda()) == {'n -> 2000',
                                           '<r,r> -> last',
                                           'm -> date_column:year'}
        tokens = [Token(x) for x in ['what', 'was', 'the', 'difference', 'in', 'attendance',
                                     'between', 'years', '2001', 'and', '2005', '?']]
        world = self._get_world_with_question_tokens(tokens)
        # "year" column does not match because "years" occurs in the question.
        assert set(world.get_agenda()) == {'n -> 2001',
                                           'n -> 2005',
                                           '<r,<r,<f,n>>> -> diff'}
        tokens = [Token(x) for x in ['what', 'was', 'the', 'total', 'avg.', 'attendance', 'in',
                                     'years', '2001', 'and', '2005', '?']]
        world = self._get_world_with_question_tokens(tokens)
        assert set(world.get_agenda()) == {'n -> 2001',
                                           'n -> 2005',
                                           '<r,<f,n>> -> sum',
                                           'f -> number_column:avg_attendance'}
        tokens = [Token(x) for x in ['when', 'was', 'the', 'least', 'avg.', 'attendance', '?']]
        world = self._get_world_with_question_tokens(tokens)
        assert set(world.get_agenda()) == {'<r,<c,r>> -> argmin', 'f -> number_column:avg_attendance'}
        tokens = [Token(x) for x in ['what', 'is', 'the', 'least', 'avg.', 'attendance', '?']]
        world = self._get_world_with_question_tokens(tokens)
        assert set(world.get_agenda()) == {'<r,<f,n>> -> min', 'f -> number_column:avg_attendance'}
        tokens = [Token(x) for x in ['when', 'did', 'the', 'team', 'not', 'qualify', '?']]
        world = self._get_world_with_question_tokens(tokens)
        assert set(world.get_agenda()) == {'s -> string:qualify'}
        tokens = [Token(x) for x in ['when', 'was', 'the', 'avg.', 'attendance', 'at', 'least',
                                     '7000', '?']]
        world = self._get_world_with_question_tokens(tokens)
        assert set(world.get_agenda()) == {'<r,<f,<n,r>>> -> filter_number_greater_equals',
                                           'f -> number_column:avg_attendance', 'n -> 7000'}
        tokens = [Token(x) for x in ['when', 'was', 'the', 'avg.', 'attendance', 'more', 'than',
                                     '7000', '?']]
        world = self._get_world_with_question_tokens(tokens)
        assert set(world.get_agenda()) == {'<r,<f,<n,r>>> -> filter_number_greater',
                                           'f -> number_column:avg_attendance', 'n -> 7000'}
        tokens = [Token(x) for x in ['when', 'was', 'the', 'avg.', 'attendance', 'at', 'most',
                                     '7000', '?']]
        world = self._get_world_with_question_tokens(tokens)
        assert set(world.get_agenda()) == {'<r,<f,<n,r>>> -> filter_number_lesser_equals',
                                           'f -> number_column:avg_attendance', 'n -> 7000'}
        tokens = [Token(x) for x in ['what', 'was', 'the', 'top', 'year', '?']]
        world = self._get_world_with_question_tokens(tokens)
        assert set(world.get_agenda()) == {'<r,r> -> first', 'm -> date_column:year'}
        tokens = [Token(x) for x in ['what', 'was', 'the', 'year', 'in', 'the', 'bottom', 'row',
                                     '?']]
        world = self._get_world_with_question_tokens(tokens)
        assert set(world.get_agenda()) == {'<r,r> -> last', 'm -> date_column:year'}



#! /usr/bin/env python
"""
Helper script for modifying config.json files that are locked inside
model.tar.gz archives. This is useful if you need to rename things or
add or remove values, usually because of changes to the library.

This script will untar the archive to a temp directory, launch an editor
to modify the config.json, and then re-tar everything to a new archive.
If your $EDITOR environment variable is not set, you'll have to explicitly
specify which editor to use.
"""
# pylint: disable=invalid-name,redefined-outer-name
import argparse
import atexit
import logging
import os
import shutil
import subprocess
import tempfile
import tarfile

from allennlp.common.file_utils import cached_path
from allennlp.models.archival import CONFIG_NAME

logger = logging.getLogger()
logger.setLevel(logging.ERROR)

def main():
    parser = argparse.ArgumentParser(description="Perform surgery on a model.tar.gz archive")

    parser.add_argument("--input-file", required=True)
    parser.add_argument("--output-file", required=True)
    parser.add_argument("--editor")

    args = parser.parse_args()

    editor = args.editor or os.environ.get("EDITOR")
    if editor is None:
        raise RuntimeError("please specify an editor or set the $EDITOR environment variable")

    if os.path.exists(args.output_file):
        raise ValueError("output file already exists")

    archive_file = cached_path(args.input_file)
    if not os.path.exists(archive_file):
        raise ValueError("input file doesn't exist")

    # Extract archive to temp dir
    tempdir = tempfile.mkdtemp()
    with tarfile.open(archive_file, 'r:gz') as archive:
        archive.extractall(tempdir)
    atexit.register(lambda: shutil.rmtree(tempdir))

    config_path = os.path.join(tempdir, CONFIG_NAME)
    subprocess.run([editor, config_path])

    with tarfile.open(args.output_file, "w:gz") as tar:
        tar.add(tempdir, arcname=os.path.sep)


if __name__ == "__main__":
    main()



# pylint: disable=no-self-use,invalid-name,too-many-public-methods
from allennlp.common.testing import AllenNlpTestCase
from allennlp.data.tokenizers import WordTokenizer
from allennlp.semparse.worlds.world import ExecutionError
from allennlp.semparse.contexts import TableQuestionContext
from allennlp.semparse.executors import WikiTablesVariableFreeExecutor
from allennlp.semparse.executors.wikitables_variable_free_executor import Date


class TestWikiTablesVariableFreeExecutor(AllenNlpTestCase):
    def setUp(self):
        super().setUp()
        table_data = [{"date_column:date": "january_2001", "number_column:division": "2",
                       "string_column:league": "usl_a_league", "string_column:regular_season": "4th_western",
                       "string_column:playoffs": "quarterfinals", "string_column:open_cup": "did_not_qualify",
                       "number_column:avg_attendance": "7169"},
                      {"date_column:date": "march_2005", "number_column:division": "2",
                       "string_column:league": "usl_first_division", "string_column:regular_season": "5th",
                       "string_column:playoffs": "quarterfinals", "string_column:open_cup": "4th_round",
                       "number_column:avg_attendance": "6028"}]
        self.executor = WikiTablesVariableFreeExecutor(table_data)

    def test_execute_fails_with_unknown_function(self):
        logical_form = "(unknown_function all_rows string_column:league)"
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_select(self):
        logical_form = "(select all_rows string_column:league)"
        cell_list = self.executor.execute(logical_form)
        assert set(cell_list) == {'usl_a_league', 'usl_first_division'}

    def test_execute_works_with_argmax(self):
        logical_form = "(select (argmax all_rows number_column:avg_attendance) string_column:league)"
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ['usl_a_league']

    def test_execute_works_with_argmax_on_dates(self):
        logical_form = "(select (argmax all_rows date_column:date) string_column:league)"
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ['usl_first_division']

    def test_execute_works_with_argmin(self):
        logical_form = "(select (argmin all_rows number_column:avg_attendance) date_column:date)"
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ['march_2005']

    def test_execute_works_with_argmin_on_dates(self):
        logical_form = "(select (argmin all_rows date_column:date) string_column:league)"
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ['usl_a_league']

    def test_execute_works_with_filter_number_greater(self):
        # Selecting cell values from all rows that have attendance greater than the min value of
        # attendance.
        logical_form = """(select (filter_number_greater all_rows number_column:avg_attendance
                                   (min all_rows number_column:avg_attendance)) string_column:league)"""
        cell_value_list = self.executor.execute(logical_form)
        assert cell_value_list == ['usl_a_league']
        # Replacing the filter value with an invalid value.
        logical_form = """(select (filter_number_greater all_rows number_column:avg_attendance
                                   all_rows) string_column:league)"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_date_greater(self):
        # Selecting cell values from all rows that have date greater than 2002.
        logical_form = """(select (filter_date_greater all_rows date_column:date
                                   (date 2002 -1 -1)) string_column:league)"""
        cell_value_list = self.executor.execute(logical_form)
        assert cell_value_list == ['usl_first_division']
        # Replacing the filter value with an invalid value.
        logical_form = """(select (filter_date_greater all_rows date_column:date
                                   2005) string_column:league)"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_number_greater_equals(self):
        # Counting rows that have attendance greater than or equal to the min value of attendance.
        logical_form = """(count (filter_number_greater_equals all_rows number_column:avg_attendance
                                  (min all_rows number_column:avg_attendance)))"""
        count_result = self.executor.execute(logical_form)
        assert count_result == 2
        # Replacing the filter value with an invalid value.
        logical_form = """(count (filter_number_greater all_rows number_column:avg_attendance all_rows))"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_date_greater_equals(self):
        # Selecting cell values from all rows that have date greater than or equal to 2005 February
        # 1st.
        logical_form = """(select (filter_date_greater_equals all_rows date_column:date
                                   (date 2005 2 1)) string_column:league)"""
        cell_value_list = self.executor.execute(logical_form)
        assert cell_value_list == ['usl_first_division']
        # Replacing the filter value with an invalid value.
        logical_form = """(select (filter_date_greater_equals all_rows date_column:date
                                   2005) string_column:league)"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_number_lesser(self):
        # Selecting cell values from all rows that have date lesser than 2005.
        logical_form = """(select (filter_number_lesser all_rows number_column:avg_attendance
                                    (max all_rows number_column:avg_attendance)) string_column:league)"""
        cell_value_list = self.executor.execute(logical_form)
        assert cell_value_list == ['usl_first_division']
        # Replacing the filter value with an invalid value.
        logical_form = """(select (filter_number_lesser all_rows date_column:date
                                   (date 2005 -1 -1)) string_column:league)"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_date_lesser(self):
        # Selecting cell values from all rows that have date less that 2005 January
        logical_form = """(select (filter_date_lesser all_rows date_column:date
                                   (date 2005 1 -1)) string_column:league)"""
        cell_value_list = self.executor.execute(logical_form)
        assert cell_value_list == ["usl_a_league"]
        # Replacing the filter value with an invalid value.
        logical_form = """(select (filter_date_lesser all_rows date_column:date
                                   2005) string_column:league)"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_number_lesser_equals(self):
        # Counting rows that have year lesser than or equal to 2005.
        logical_form = """(count (filter_number_lesser_equals all_rows number_column:avg_attendance 8000))"""
        count_result = self.executor.execute(logical_form)
        assert count_result == 2
        # Replacing the filter value with an invalid value.
        logical_form = """(select (filter_number_lesser_equals all_rows date_column:date
                                   (date 2005 -1 -1)) string_column:league)"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_date_lesser_equals(self):
        # Selecting cell values from all rows that have date less that or equal to 2001 February 23
        logical_form = """(select (filter_date_lesser_equals all_rows date_column:date
                                   (date 2001 2 23)) string_column:league)"""
        cell_value_list = self.executor.execute(logical_form)
        assert cell_value_list == ['usl_a_league']
        # Replacing the filter value with an invalid value.
        logical_form = """(select (filter_date_lesser_equals all_rows date_column:date
                                   2005) string_column:league)"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_number_equals(self):
        # Counting rows that have year equal to 2010.
        logical_form = """(count (filter_number_equals all_rows number_column:avg_attendance 8000))"""
        count_result = self.executor.execute(logical_form)
        assert count_result == 0
        # Replacing the filter value with an invalid value.
        logical_form = """(count (filter_number_equals all_rows date_column:date (date 2010 -1 -1)))"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_date_equals(self):
        # Selecting cell values from all rows that have date not equal to 2001
        logical_form = """(select (filter_date_equals all_rows date_column:date
                                   (date 2001 -1 -1)) string_column:league)"""
        cell_value_list = self.executor.execute(logical_form)
        assert cell_value_list == ['usl_a_league']
        # Replacing the filter value with an invalid value.
        logical_form = """(select (filter_date_equals all_rows date_column:date
                                   2005) string_column:league)"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_number_not_equals(self):
        # Counting rows that have year not equal to 2010.
        logical_form = """(count (filter_number_not_equals all_rows number_column:avg_attendance 8000))"""
        count_result = self.executor.execute(logical_form)
        assert count_result == 2
        # Replacing the filter value with an invalid value.
        logical_form = """(count (filter_number_not_equals all_rows date_column:date (date 2010 -1 -1)))"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_date_not_equals(self):
        # Selecting cell values from all rows that have date not equal to 2001
        logical_form = """(select (filter_date_not_equals all_rows date_column:date
                                   (date 2001 -1 -1)) string_column:league)"""
        cell_value_list = self.executor.execute(logical_form)
        assert cell_value_list == ['usl_first_division']
        # Replacing the filter value with an invalid value.
        logical_form = """(select (filter_date_not_equals all_rows date_column:date
                                   2005) string_column:league)"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_works_with_filter_in(self):
        # Selecting "regular season" from rows that have "did not qualify" in "open cup" column.
        logical_form = """(select (filter_in all_rows string_column:open_cup string:did_not_qualify)
                                  string_column:regular_season)"""
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ["4th_western"]

    def test_execute_works_with_filter_not_in(self):
        # Selecting "regular season" from rows that do not have "did not qualify" in "open cup" column.
        logical_form = """(select (filter_not_in all_rows string_column:open_cup string:did_not_qualify)
                                   string_column:regular_season)"""
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ["5th"]

    def test_execute_works_with_first(self):
        # Selecting "regular season" from the first row.
        logical_form = """(select (first all_rows) string_column:regular_season)"""
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ["4th_western"]

    def test_execute_logs_warning_with_first_on_empty_list(self):
        # Selecting "regular season" from the first row where year is greater than 2010.
        with self.assertLogs("allennlp.semparse.executors.wikitables_variable_free_executor") as log:
            logical_form = """(select (first (filter_date_greater all_rows date_column:date
                                                (date 2010 -1 -1)))
                                      string_column:regular_season)"""
            self.executor.execute(logical_form)
        self.assertEqual(log.output,
                         ["WARNING:allennlp.semparse.executors.wikitables_variable_free_executor:"
                          "Trying to get first row from an empty list: "
                          "['filter_date_greater', 'all_rows', 'date_column:date', ['date', '2010', '-1', '-1']]"])

    def test_execute_works_with_last(self):
        # Selecting "regular season" from the last row where year is not equal to 2010.
        logical_form = """(select (last (filter_date_not_equals all_rows date_column:date
                                         (date 2010 -1 -1)))
                                  string_column:regular_season)"""
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ["5th"]

    def test_execute_logs_warning_with_last_on_empty_list(self):
        # Selecting "regular season" from the last row where year is greater than 2010.
        with self.assertLogs("allennlp.semparse.executors.wikitables_variable_free_executor") as log:
            logical_form = """(select (last (filter_date_greater all_rows date_column:date
                                                (date 2010 -1 -1)))
                                      string_column:regular_season)"""
            self.executor.execute(logical_form)
        self.assertEqual(log.output,
                         ["WARNING:allennlp.semparse.executors.wikitables_variable_free_executor:"
                          "Trying to get last row from an empty list: "
                          "['filter_date_greater', 'all_rows', 'date_column:date', ['date', '2010', '-1', '-1']]"])

    def test_execute_works_with_previous(self):
        # Selecting "regular season" from the row before last where year is not equal to 2010.
        logical_form = """(select (previous (last (filter_date_not_equals
                                                    all_rows date_column:date (date 2010 -1 -1))))
                                  string_column:regular_season)"""
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ["4th_western"]

    def test_execute_logs_warning_with_previous_on_empty_list(self):
        # Selecting "regular season" from the row before the one where year is greater than 2010.
        with self.assertLogs("allennlp.semparse.executors.wikitables_variable_free_executor") as log:
            logical_form = """(select (previous (filter_date_greater all_rows date_column:date (date 2010 -1 -1)))
                                      string_column:regular_season)"""
            self.executor.execute(logical_form)
        self.assertEqual(log.output,
                         ["WARNING:allennlp.semparse.executors.wikitables_variable_free_executor:"
                          "Trying to get the previous row from an empty list: "
                          "['filter_date_greater', 'all_rows', 'date_column:date', ['date', '2010', '-1', '-1']]"])

    def test_execute_works_with_next(self):
        # Selecting "regular season" from the row after first where year is not equal to 2010.
        logical_form = """(select (next (first (filter_date_not_equals
                                                all_rows date_column:date (date 2010 -1 -1))))
                                  string_column:regular_season)"""
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ["5th"]

    def test_execute_logs_warning_with_next_on_empty_list(self):
        # Selecting "regular season" from the row after the one where year is greater than 2010.
        with self.assertLogs("allennlp.semparse.executors.wikitables_variable_free_executor") as log:
            logical_form = """(select (next (filter_date_greater all_rows date_column:date (date 2010 -1 -1)))
                                      string_column:regular_season)"""
            self.executor.execute(logical_form)
        self.assertEqual(log.output,
                         ["WARNING:allennlp.semparse.executors.wikitables_variable_free_executor:"
                          "Trying to get the next row from an empty list: "
                          "['filter_date_greater', 'all_rows', 'date_column:date', ['date', '2010', '-1', '-1']]"])

    def test_execute_works_with_mode(self):
        # Most frequent division value.
        logical_form = """(mode all_rows number_column:division)"""
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ["2"]
        # If we used select instead, we should get a list of two values.
        logical_form = """(select all_rows number_column:division)"""
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ["2", "2"]
        # If we queried for the most frequent year instead, it should return two values since both
        # have the max frequency of 1.
        logical_form = """(mode all_rows date_column:date)"""
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ["january_2001", "march_2005"]

    def test_execute_works_with_same_as(self):
        # Select the "league" from all the rows that have the same value under "playoffs" as the
        # row that has the string "a league" under "league".
        logical_form = """(select (same_as (filter_in all_rows string_column:league string:a_league)
                                   string_column:playoffs)
                           string_column:league)"""
        cell_list = self.executor.execute(logical_form)
        assert cell_list == ["usl_a_league", "usl_first_division"]

    def test_execute_works_with_sum(self):
        # Get total "avg attendance".
        logical_form = """(sum all_rows number_column:avg_attendance)"""
        sum_value = self.executor.execute(logical_form)
        assert sum_value == 13197
        # Total "avg attendance" where "playoffs" has "quarterfinals"
        logical_form = """(sum (filter_in all_rows string_column:playoffs string:quarterfinals)
                                number_column:avg_attendance)"""
        sum_value = self.executor.execute(logical_form)
        assert sum_value == 13197

    def test_execute_works_with_average(self):
        # Get average "avg attendance".
        logical_form = """(average all_rows number_column:avg_attendance)"""
        avg_value = self.executor.execute(logical_form)
        assert avg_value == 6598.5
        # Average "avg attendance" where "playoffs" has "quarterfinals"
        logical_form = """(average (filter_in all_rows string_column:playoffs string:quarterfinals)
                                number_column:avg_attendance)"""
        avg_value = self.executor.execute(logical_form)
        assert avg_value == 6598.5

    def test_execute_works_with_diff(self):
        # Difference in "avg attendance" between rows with "usl_a_league" and "usl_first_division"
        # in "league" columns.
        logical_form = """(diff (filter_in all_rows string_column:league string:usl_a_league)
                                (filter_in all_rows string_column:league string:usl_first_division)
                                number_column:avg_attendance)"""
        avg_value = self.executor.execute(logical_form)
        assert avg_value == 1141

    def test_execute_fails_with_diff_on_non_numerical_columns(self):
        logical_form = """(diff (filter_in all_rows string_column:league string:usl_a_league)
                                (filter_in all_rows string_column:league string:usl_first_division)
                                string_column:league)"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_execute_fails_with_non_int_dates(self):
        logical_form = """(date 2015 1.5 1)"""
        with self.assertRaises(ExecutionError):
            self.executor.execute(logical_form)

    def test_date_comparison_works(self):
        assert Date(2013, 12, 31) > Date(2013, 12, 30)
        assert Date(2013, 12, 31) == Date(2013, 12, -1)
        assert Date(2013, -1, -1) >= Date(2013, 12, 31)
        # pylint: disable=singleton-comparison
        assert (Date(2013, 12, -1) > Date(2013, 12, 31)) == False
        assert (Date(2013, 12, 31) > 2013) == False
        assert (Date(2013, 12, 31) >= 2013) == False
        assert Date(2013, 12, 31) != 2013
        assert (Date(2018, 1, 1) >= Date(-1, 2, 1)) == False
        assert (Date(2018, 1, 1) < Date(-1, 2, 1)) == False
        # When year is unknown in both cases, we can compare months and days.
        assert Date(-1, 2, 1) < Date(-1, 2, 3)
        # If both year and month are not know in both cases, the comparison is undefined, and both
        # < and >= return False.
        assert (Date(-1, -1, 1) < Date(-1, -1, 3)) == False
        assert (Date(-1, -1, 1) >= Date(-1, -1, 3)) == False
        # Same when year is known, buth months are not.
        assert (Date(2018, -1, 1) < Date(2018, -1, 3)) == False
        # TODO (pradeep): Figure out whether this is expected behavior by looking at data.
        assert (Date(2018, -1, 1) >= Date(2018, -1, 3)) == False

    def test_number_comparison_works(self):
        # TableQuestionContext normlaizes all strings according to some rules. We want to ensure
        # that the original numerical values of number cells is being correctly processed here.
        tokens = WordTokenizer().tokenize("when was the attendance the highest?")
        tagged_file = self.FIXTURES_ROOT / "data" / "corenlp_processed_tables" / "TEST-2.table"
        context = TableQuestionContext.read_from_file(tagged_file, tokens)
        executor = WikiTablesVariableFreeExecutor(context.table_data)
        result = executor.execute("(select (argmax all_rows number_column:attendance) date_column:date)")
        assert result == ["november_10"]

    def test_evaluate_logical_form(self):
        logical_form = """(select (same_as (filter_in all_rows string_column:league string:a_league)
                                   string_column:playoffs)
                           string_column:league)"""
        assert self.executor.evaluate_logical_form(logical_form, ["USL A-League",
                                                                  "USL First Division"])

    def test_evaluate_logical_form_with_invalid_logical_form(self):
        logical_form = """(select (same_as (filter_in all_rows string_column:league INVALID_CONSTANT)
                                   string_column:playoffs)
                           string_column:league)"""
        assert not self.executor.evaluate_logical_form(logical_form, ["USL A-League",
                                                                      "USL First Division"])



# pylint: disable=invalid-name,too-many-public-methods
import glob
import json
import os
import re
import time
from typing import Dict

import torch
import pytest
from allennlp.common.checks import ConfigurationError

from allennlp.common.testing import AllenNlpTestCase
from allennlp.training.trainer import Trainer, sparse_clip_norm, is_sparse
from allennlp.training.learning_rate_schedulers import LearningRateScheduler
from allennlp.data import Vocabulary
from allennlp.common.params import Params
from allennlp.models.simple_tagger import SimpleTagger
from allennlp.data.iterators import BasicIterator
from allennlp.data.dataset_readers import SequenceTaggingDatasetReader
from allennlp.models.model import Model


class TestTrainer(AllenNlpTestCase):
    def setUp(self):
        super(TestTrainer, self).setUp()
        self.instances = SequenceTaggingDatasetReader().read(self.FIXTURES_ROOT / 'data' / 'sequence_tagging.tsv')
        vocab = Vocabulary.from_instances(self.instances)
        self.vocab = vocab
        self.model_params = Params({
                "text_field_embedder": {
                        "token_embedders": {
                                "tokens": {
                                        "type": "embedding",
                                        "embedding_dim": 5
                                        }
                                }
                        },
                "encoder": {
                        "type": "lstm",
                        "input_size": 5,
                        "hidden_size": 7,
                        "num_layers": 2
                        }
                })
        self.model = SimpleTagger.from_params(vocab=self.vocab, params=self.model_params)
        self.optimizer = torch.optim.SGD(self.model.parameters(), 0.01)
        self.iterator = BasicIterator(batch_size=2)
        self.iterator.index_with(vocab)

    def test_trainer_can_run(self):
        trainer = Trainer(model=self.model,
                          optimizer=self.optimizer,
                          iterator=self.iterator,
                          train_dataset=self.instances,
                          validation_dataset=self.instances,
                          num_epochs=2)
        metrics = trainer.train()
        assert 'best_validation_loss' in metrics
        assert isinstance(metrics['best_validation_loss'], float)
        assert 'best_validation_accuracy' in metrics
        assert isinstance(metrics['best_validation_accuracy'], float)
        assert 'best_validation_accuracy3' in metrics
        assert isinstance(metrics['best_validation_accuracy3'], float)
        assert 'best_epoch' in metrics
        assert isinstance(metrics['best_epoch'], int)

        # Making sure that both increasing and decreasing validation metrics work.
        trainer = Trainer(model=self.model,
                          optimizer=self.optimizer,
                          iterator=self.iterator,
                          train_dataset=self.instances,
                          validation_dataset=self.instances,
                          validation_metric='+loss',
                          num_epochs=2)
        metrics = trainer.train()
        assert 'best_validation_loss' in metrics
        assert isinstance(metrics['best_validation_loss'], float)
        assert 'best_validation_accuracy' in metrics
        assert isinstance(metrics['best_validation_accuracy'], float)
        assert 'best_validation_accuracy3' in metrics
        assert isinstance(metrics['best_validation_accuracy3'], float)
        assert 'best_epoch' in metrics
        assert isinstance(metrics['best_epoch'], int)
        assert 'peak_cpu_memory_MB' in metrics
        assert isinstance(metrics['peak_cpu_memory_MB'], float)
        assert metrics['peak_cpu_memory_MB'] > 0

    @pytest.mark.skipif(not torch.cuda.is_available(), reason="No CUDA device registered.")
    def test_trainer_can_run_cuda(self):
        trainer = Trainer(self.model, self.optimizer,
                          self.iterator, self.instances, num_epochs=2,
                          cuda_device=0)
        trainer.train()

    @pytest.mark.skipif(torch.cuda.device_count() < 2,
                        reason="Need multiple GPUs.")
    def test_trainer_can_run_multiple_gpu(self):

        class MetaDataCheckWrapper(Model):
            """
            Checks that the metadata field has been correctly split across the batch dimension
            when running on multiple gpus.
            """
            def __init__(self, model):
                super().__init__(model.vocab)
                self.model = model

            def forward(self, **kwargs) -> Dict[str, torch.Tensor]:  # type: ignore # pylint: disable=arguments-differ
                assert 'metadata' in kwargs and 'tags' in kwargs, \
                    f'tokens and metadata must be provided. Got {kwargs.keys()} instead.'
                batch_size = kwargs['tokens']['tokens'].size()[0]
                assert len(kwargs['metadata']) == batch_size, \
                    f'metadata must be split appropriately. Expected {batch_size} elements, ' \
                    f"got {len(kwargs['metadata'])} elements."
                return self.model.forward(**kwargs)

        multigpu_iterator = BasicIterator(batch_size=4)
        multigpu_iterator.index_with(self.vocab)
        trainer = Trainer(MetaDataCheckWrapper(self.model), self.optimizer,
                          multigpu_iterator, self.instances, num_epochs=2,
                          cuda_device=[0, 1])
        metrics = trainer.train()
        assert 'peak_cpu_memory_MB' in metrics
        assert isinstance(metrics['peak_cpu_memory_MB'], float)
        assert metrics['peak_cpu_memory_MB'] > 0
        assert 'peak_gpu_0_memory_MB' in metrics
        assert isinstance(metrics['peak_gpu_0_memory_MB'], int)
        assert 'peak_gpu_1_memory_MB' in metrics
        assert isinstance(metrics['peak_gpu_1_memory_MB'], int)

    def test_trainer_can_resume_training(self):
        trainer = Trainer(self.model, self.optimizer,
                          self.iterator, self.instances,
                          validation_dataset=self.instances,
                          num_epochs=1, serialization_dir=self.TEST_DIR)
        trainer.train()
        new_trainer = Trainer(self.model, self.optimizer,
                              self.iterator, self.instances,
                              validation_dataset=self.instances,
                              num_epochs=3, serialization_dir=self.TEST_DIR)

        epoch, val_metrics_per_epoch = new_trainer._restore_checkpoint()  # pylint: disable=protected-access
        assert epoch == 1
        assert len(val_metrics_per_epoch) == 1
        assert isinstance(val_metrics_per_epoch[0], float)
        assert val_metrics_per_epoch[0] != 0.
        new_trainer.train()

    def test_metric_only_considered_best_so_far_when_strictly_better_than_those_before_it_increasing_metric(
            self):
        new_trainer = Trainer(self.model, self.optimizer,
                              self.iterator, self.instances,
                              validation_dataset=self.instances,
                              num_epochs=3, serialization_dir=self.TEST_DIR,
                              patience=5, validation_metric="+test")
        # when it is the only metric it should be considered the best
        assert new_trainer._is_best_so_far(1, [])  # pylint: disable=protected-access
        # when it is the same as one before it it is not considered the best
        assert not new_trainer._is_best_so_far(.3, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access
        # when it is the best it is considered the best
        assert new_trainer._is_best_so_far(13.00, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access
        # when it is not the the best it is not considered the best
        assert not new_trainer._is_best_so_far(.0013, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access

    def test_metric_only_considered_best_so_far_when_strictly_better_than_those_before_it_decreasing_metric(self):
        new_trainer = Trainer(self.model, self.optimizer,
                              self.iterator, self.instances,
                              validation_dataset=self.instances,
                              num_epochs=3, serialization_dir=self.TEST_DIR,
                              patience=5, validation_metric="-test")
        # when it is the only metric it should be considered the best
        assert new_trainer._is_best_so_far(1, [])  # pylint: disable=protected-access
        # when it is the same as one before it it is not considered the best
        assert not new_trainer._is_best_so_far(.3, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access
        # when it is the best it is considered the best
        assert new_trainer._is_best_so_far(.013, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access
        # when it is not the the best it is not considered the best
        assert not new_trainer._is_best_so_far(13.00, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access

    def test_should_stop_early_with_increasing_metric(self):
        new_trainer = Trainer(self.model, self.optimizer,
                              self.iterator, self.instances,
                              validation_dataset=self.instances,
                              num_epochs=3, serialization_dir=self.TEST_DIR,
                              patience=5, validation_metric="+test")
        assert new_trainer._should_stop_early([.5, .3, .2, .1, .4, .4])  # pylint: disable=protected-access
        assert not new_trainer._should_stop_early([.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access

    def test_should_stop_early_with_flat_lining_metric(self):
        flatline = [.2] * 6
        assert Trainer(self.model, self.optimizer,  # pylint: disable=protected-access
                       self.iterator, self.instances,
                       validation_dataset=self.instances,
                       num_epochs=3,
                       serialization_dir=self.TEST_DIR,
                       patience=5,
                       validation_metric="+test")._should_stop_early(flatline)  # pylint: disable=protected-access
        assert Trainer(self.model, self.optimizer,  # pylint: disable=protected-access
                       self.iterator, self.instances,
                       validation_dataset=self.instances,
                       num_epochs=3,
                       serialization_dir=self.TEST_DIR,
                       patience=5,
                       validation_metric="-test")._should_stop_early(flatline)  # pylint: disable=protected-access

    def test_should_stop_early_with_decreasing_metric(self):
        new_trainer = Trainer(self.model, self.optimizer,
                              self.iterator, self.instances,
                              validation_dataset=self.instances,
                              num_epochs=3, serialization_dir=self.TEST_DIR,
                              patience=5, validation_metric="-test")
        assert new_trainer._should_stop_early([.02, .3, .2, .1, .4, .4])  # pylint: disable=protected-access
        assert not new_trainer._should_stop_early([.3, .3, .2, .1, .4, .5])  # pylint: disable=protected-access
        assert new_trainer._should_stop_early([.1, .3, .2, .1, .4, .5])  # pylint: disable=protected-access

    def test_should_stop_early_with_early_stopping_disabled(self):
        # Increasing metric
        trainer = Trainer(self.model, self.optimizer, self.iterator, self.instances,
                          validation_dataset=self.instances, num_epochs=100,
                          patience=None, validation_metric="+test")
        decreasing_history = [float(i) for i in reversed(range(20))]
        assert not trainer._should_stop_early(decreasing_history)  # pylint: disable=protected-access

        # Decreasing metric
        trainer = Trainer(self.model, self.optimizer, self.iterator, self.instances,
                          validation_dataset=self.instances, num_epochs=100,
                          patience=None, validation_metric="-test")
        increasing_history = [float(i) for i in range(20)]
        assert not trainer._should_stop_early(increasing_history)  # pylint: disable=protected-access

    def test_should_stop_early_with_invalid_patience(self):
        for patience in [0, -1, -2, 1.5, 'None']:
            with pytest.raises(ConfigurationError,
                               message='No ConfigurationError for patience={}'.format(patience)):
                Trainer(self.model, self.optimizer, self.iterator, self.instances,
                        validation_dataset=self.instances, num_epochs=100,
                        patience=patience, validation_metric="+test")

    def test_trainer_can_run_with_lr_scheduler(self):
        lr_params = Params({"type": "reduce_on_plateau"})
        lr_scheduler = LearningRateScheduler.from_params(self.optimizer, lr_params)
        trainer = Trainer(model=self.model,
                          optimizer=self.optimizer,
                          iterator=self.iterator,
                          learning_rate_scheduler=lr_scheduler,
                          validation_metric="-loss",
                          train_dataset=self.instances,
                          validation_dataset=self.instances,
                          num_epochs=2)
        trainer.train()

    def test_trainer_can_resume_with_lr_scheduler(self):
        # pylint: disable=protected-access
        lr_scheduler = LearningRateScheduler.from_params(
                self.optimizer, Params({"type": "exponential", "gamma": 0.5}))
        trainer = Trainer(model=self.model,
                          optimizer=self.optimizer,
                          iterator=self.iterator,
                          learning_rate_scheduler=lr_scheduler,
                          train_dataset=self.instances,
                          validation_dataset=self.instances,
                          num_epochs=2, serialization_dir=self.TEST_DIR)
        trainer.train()

        new_lr_scheduler = LearningRateScheduler.from_params(
                self.optimizer, Params({"type": "exponential", "gamma": 0.5}))
        new_trainer = Trainer(model=self.model,
                              optimizer=self.optimizer,
                              iterator=self.iterator,
                              learning_rate_scheduler=new_lr_scheduler,
                              train_dataset=self.instances,
                              validation_dataset=self.instances,
                              num_epochs=4, serialization_dir=self.TEST_DIR)
        epoch, _ = new_trainer._restore_checkpoint()
        assert epoch == 2
        assert new_trainer._learning_rate_scheduler.lr_scheduler.last_epoch == 1
        new_trainer.train()

    def test_trainer_raises_on_model_with_no_loss_key(self):
        class FakeModel(torch.nn.Module):
            def forward(self, **kwargs):  # pylint: disable=arguments-differ,unused-argument
                return {}
        with pytest.raises(RuntimeError):
            trainer = Trainer(FakeModel(), self.optimizer,
                              self.iterator, self.instances,
                              num_epochs=2, serialization_dir=self.TEST_DIR)
            trainer.train()

    def test_trainer_can_log_histograms(self):
        # enable activation logging
        for module in self.model.modules():
            module.should_log_activations = True

        trainer = Trainer(self.model, self.optimizer,
                          self.iterator, self.instances, num_epochs=3,
                          serialization_dir=self.TEST_DIR,
                          histogram_interval=2)
        trainer.train()

    def test_trainer_respects_num_serialized_models_to_keep(self):
        trainer = Trainer(self.model, self.optimizer,
                          self.iterator, self.instances, num_epochs=5,
                          serialization_dir=self.TEST_DIR,
                          num_serialized_models_to_keep=3)
        trainer.train()

        # Now check the serialized files
        for prefix in ['model_state_epoch_*', 'training_state_epoch_*']:
            file_names = glob.glob(os.path.join(self.TEST_DIR, prefix))
            epochs = [int(re.search(r"_([0-9])\.th", fname).group(1))
                      for fname in file_names]
            assert sorted(epochs) == [2, 3, 4]

    def test_trainer_saves_metrics_every_epoch(self):
        trainer = Trainer(model=self.model,
                          optimizer=self.optimizer,
                          iterator=self.iterator,
                          train_dataset=self.instances,
                          validation_dataset=self.instances,
                          num_epochs=5,
                          serialization_dir=self.TEST_DIR,
                          num_serialized_models_to_keep=3)
        trainer.train()

        for epoch in range(5):
            epoch_file = self.TEST_DIR / f'metrics_epoch_{epoch}.json'
            assert epoch_file.exists()
            metrics = json.load(open(epoch_file))
            assert "validation_loss" in metrics
            assert "best_validation_loss" in metrics
            assert metrics.get("epoch") == epoch

    def test_trainer_respects_keep_serialized_model_every_num_seconds(self):
        # To test:
        #   Create an iterator that sleeps for 2.5 second per epoch, so the total training
        #       time for one epoch is slightly greater then 2.5 seconds.
        #   Run for 6 epochs, keeping the last 2 models, models also kept every 5 seconds.
        #   Check the resulting checkpoints.  Should then have models at epochs
        #       2, 4, plus the last two at 5 and 6.
        class WaitingIterator(BasicIterator):
            # pylint: disable=arguments-differ
            def _create_batches(self, *args, **kwargs):
                time.sleep(2.5)
                return super(WaitingIterator, self)._create_batches(*args, **kwargs)

        iterator = WaitingIterator(batch_size=2)
        iterator.index_with(self.vocab)

        trainer = Trainer(self.model, self.optimizer,
                          iterator, self.instances, num_epochs=6,
                          serialization_dir=self.TEST_DIR,
                          num_serialized_models_to_keep=2,
                          keep_serialized_model_every_num_seconds=5)
        trainer.train()

        # Now check the serialized files
        for prefix in ['model_state_epoch_*', 'training_state_epoch_*']:
            file_names = glob.glob(os.path.join(self.TEST_DIR, prefix))
            epochs = [int(re.search(r"_([0-9])\.th", fname).group(1))
                      for fname in file_names]
            # epoch N has N-1 in file name
            assert sorted(epochs) == [1, 3, 4, 5]

    def test_trainer_can_log_learning_rates_tensorboard(self):
        iterator = BasicIterator(batch_size=4)
        iterator.index_with(self.vocab)

        trainer = Trainer(self.model, self.optimizer,
                          iterator, self.instances, num_epochs=2,
                          serialization_dir=self.TEST_DIR,
                          should_log_learning_rate=True,
                          summary_interval=2)

        trainer.train()

    def test_trainer_saves_models_at_specified_interval(self):
        iterator = BasicIterator(batch_size=4)
        iterator.index_with(self.vocab)

        trainer = Trainer(self.model, self.optimizer,
                          iterator, self.instances, num_epochs=2,
                          serialization_dir=self.TEST_DIR,
                          model_save_interval=0.0001)

        trainer.train()

        # Now check the serialized files for models saved during the epoch.
        prefix = 'model_state_epoch_*'
        file_names = sorted(glob.glob(os.path.join(self.TEST_DIR, prefix)))
        epochs = [re.search(r"_([0-9\.\-]+)\.th", fname).group(1)
                  for fname in file_names]
        # We should have checkpoints at the end of each epoch and during each, e.g.
        # [0.timestamp, 0, 1.timestamp, 1]
        assert len(epochs) == 4
        assert epochs[3] == '1'
        assert '.' in epochs[0]

        # Now make certain we can restore from timestamped checkpoint.
        # To do so, remove the checkpoint from the end of epoch 1&2, so
        # that we are forced to restore from the timestamped checkpoints.
        for k in range(2):
            os.remove(os.path.join(self.TEST_DIR, 'model_state_epoch_{}.th'.format(k)))
            os.remove(os.path.join(self.TEST_DIR, 'training_state_epoch_{}.th'.format(k)))
        os.remove(os.path.join(self.TEST_DIR, 'best.th'))

        restore_trainer = Trainer(self.model, self.optimizer,
                                  self.iterator, self.instances, num_epochs=2,
                                  serialization_dir=self.TEST_DIR,
                                  model_save_interval=0.0001)
        epoch, _ = restore_trainer._restore_checkpoint()  # pylint: disable=protected-access
        assert epoch == 2
        # One batch per epoch.
        assert restore_trainer._batch_num_total == 2  # pylint: disable=protected-access


class TestSparseClipGrad(AllenNlpTestCase):
    def test_sparse_clip_grad(self):
        # create a sparse embedding layer, then take gradient
        embedding = torch.nn.Embedding(100, 16, sparse=True)
        embedding.zero_grad()
        ids = (torch.rand(17) * 100).long()
        # Set some of the ids to the same value so that the sparse gradient
        # has repeated indices.  This tests some additional logic.
        ids[:5] = 5
        loss = embedding(ids).sum()
        loss.backward()
        assert is_sparse(embedding.weight.grad)

        # Now try to clip the gradients.
        _ = sparse_clip_norm([embedding.weight], 1.5)
        # Final norm should be 1.5
        grad = embedding.weight.grad.coalesce()  # pylint: disable=no-member
        self.assertAlmostEqual(grad._values().norm(2.0).item(), 1.5, places=5) # pylint: disable=protected-access
